from requests import Response

from yawast.shared import utils


def is_text(res: Response) -> bool:
    """
    Returns True if the body is HTML, or at least seems like text
    :param res:
    :return:
    """
    has_text = False

    if len(res.content) == 0:
        # don't bother with these, if the body is empty
        has_text = False
    elif "Content-Type" in res.headers and "text/html" in res.headers["Content-Type"]:
        # it's HTML, go
        has_text = True
    elif "Content-Type" not in res.headers:
        # this is something, but the server doesn't tell us what
        # so, we will check to see if if we can treat it like text
        if utils.is_printable_str(res.content):
            has_text = True

    return has_text



import secrets
from http import cookiejar
from typing import List, Dict, Union, Tuple
from urllib.parse import urlparse, urljoin
from urllib.parse import urlunparse

import requests
import urllib3
from requests.adapters import HTTPAdapter
from requests.models import Response, Request
from urllib3 import Retry

from yawast._version import get_version
from yawast.shared import output

YAWAST_UA = (
    f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_5) AppleWebKit/537.36 (KHTML, like Gecko) "
    f"YAWAST/{get_version()}/PY Chrome/74.0.3729.169 Safari/537.36"
)

SERVICE_UA = f"YAWAST/{get_version()}/PY"


# class to block setting cookies from server responses
class _BlockCookiesSet(cookiejar.DefaultCookiePolicy):
    def set_ok(self, cookie, request):
        return False


_requester = requests.Session()


def init(proxy: str, cookie: str) -> None:
    global _requester

    _requester.cookies.set_policy(_BlockCookiesSet())
    _requester.mount(
        "http://",
        HTTPAdapter(
            max_retries=Retry(total=5, read=5, connect=5, backoff_factor=0.3),
            pool_maxsize=50,
        ),
    )
    _requester.mount(
        "https://",
        HTTPAdapter(
            max_retries=Retry(total=5, read=5, connect=5, backoff_factor=0.3),
            pool_maxsize=50,
        ),
    )

    if proxy is not None and len(proxy) > 0:
        # we have a proxy, set it
        if not proxy.startswith("http"):
            proxy = f"http://{proxy}"

        proxies = {"http": proxy, "https": proxy}

        _requester.proxies.update(proxies)

    if cookie is not None and len(cookie) > 0:
        name = cookie.split("=")[0]
        val = cookie.split("=")[1]
        c = requests.cookies.create_cookie(name=name, value=val)

        _requester.cookies.set_cookie(c)


def http_head(url, allow_redirects=True, timeout=15) -> Response:
    global _requester

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    headers = {"User-Agent": YAWAST_UA}
    res = _requester.head(
        url,
        headers=headers,
        verify=False,
        allow_redirects=allow_redirects,
        timeout=timeout,
    )

    output.debug(
        f"{res.request.method}: {url} - completed ({res.status_code}) in "
        f"{int(res.elapsed.total_seconds() * 1000)}ms."
    )

    return res


def http_options(url, timeout=15) -> Response:
    global _requester

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    headers = {"User-Agent": YAWAST_UA}
    res = _requester.options(url, headers=headers, verify=False, timeout=timeout)

    output.debug(
        f"{res.request.method}: {url} - completed ({res.status_code}) in "
        f"{int(res.elapsed.total_seconds() * 1000)}ms."
    )

    return res


def http_get(
    url: str, allow_redirects=True, additional_headers: Union[None, Dict] = None
) -> Response:

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    global _requester

    headers = {"User-Agent": YAWAST_UA}

    if additional_headers is not None:
        headers = {**headers, **additional_headers}

    res = _requester.get(
        url, headers=headers, verify=False, allow_redirects=allow_redirects
    )

    output.debug(
        f"{res.request.method}: {url} - completed ({res.status_code}) in "
        f"{int(res.elapsed.total_seconds() * 1000)}ms."
    )

    return res


def http_put(
    url: str,
    data: str,
    allow_redirects=True,
    additional_headers: Union[None, Dict] = None,
) -> Response:

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    global _requester

    headers = {"User-Agent": YAWAST_UA}

    if additional_headers is not None:
        headers = {**headers, **additional_headers}

    res = _requester.put(
        url, data=data, headers=headers, verify=False, allow_redirects=allow_redirects
    )

    output.debug(
        f"{res.request.method}: {url} - completed ({res.status_code}) in "
        f"{int(res.elapsed.total_seconds() * 1000)}ms."
    )

    return res


def http_custom(
    verb: str, url: str, additional_headers: Union[None, Dict] = None
) -> Response:

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    global _requester

    headers = {"User-Agent": YAWAST_UA}

    if additional_headers is not None:
        headers = {**headers, **additional_headers}

    res = _requester.request(verb, url, headers=headers, verify=False)

    output.debug(
        f"{res.request.method}: {url} - completed ({res.status_code}) in "
        f"{int(res.elapsed.total_seconds() * 1000)}ms."
    )

    return res


def http_json(url, allow_redirects=True) -> Tuple[Dict, int]:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    global _requester

    headers = {"User-Agent": SERVICE_UA}

    res = _requester.get(
        url, headers=headers, verify=False, allow_redirects=allow_redirects
    )
    return res.json(), res.status_code


def http_build_raw_response(res: Response) -> str:
    lines = []

    if res.raw.version == 11:
        res_line = f"HTTP/1.1 {res.raw.status} {res.raw.reason}"
    elif res.raw.version == 10:
        res_line = f"HTTP/1.0 {res.raw.status} {res.raw.reason}"
    else:
        raise ValueError(f"Invalid HTTP version ({res.raw.version})")

    if res_line != "":
        lines.append(res_line)

    for header in res.headers:
        lines.append(f"{header}: {res.headers[header]}")

    try:
        txt = res.text

        if txt != "":
            lines.append("")
            lines.append("")

            for line in txt.splitlines():
                lines.append(line)
    except Exception:
        output.debug_exception()

    return "\r\n".join(lines)


def http_build_raw_request(req: Request) -> str:
    headers = "\n".join(f"{k}: {v}" for k, v in req.headers.items())

    body = ""
    if req.body is not None:
        body = req.body

    return f"{req.method} {req.url}\n{headers}\n\n{body}"


def check_404_response(url: str) -> [bool, bool]:
    rnd = secrets.token_hex(12)
    file_url = urljoin(url, f"{rnd}.html")
    path_url = urljoin(url, f"{rnd}/")

    file_res = http_get(file_url, False)
    path_res = http_get(path_url, False)

    return file_res.status_code == 404, path_res.status_code == 404


def check_ssl_redirect(url):
    parsed = urlparse(url)

    if parsed.scheme == "https":
        return url

    req = http_head(url, False)

    # make sure we received a redirect response
    if req.status_code >= 300 & req.status_code < 400:
        location = req.headers.get("location")

        if location is None:
            return url

        try:
            parsed_location = urlparse(location)

            # this is a special case to handle servers that redirect to a path, and then to HTTPS
            if parsed_location.netloc == "" and parsed_location.path != "":
                parsed_location = parsed._replace(path=parsed_location.path)
                parsed_location = urlparse(
                    check_ssl_redirect(urlunparse(parsed_location))
                )

            if parsed_location.scheme == "https":
                parsed = parsed._replace(scheme=parsed_location.scheme)

                return urlunparse(parsed)
        except ValueError:
            return url
    else:
        return url


def check_www_redirect(url):
    parsed = urlparse(url)

    req = http_head(url, False)

    # make sure we received a redirect response
    if req.status_code >= 300 & req.status_code < 400:
        location = req.headers.get("location")

        if location is None:
            return url

        try:
            parsed_location = urlparse(location)

            if parsed.netloc.startswith("www") & (
                not parsed_location.netloc.startswith("www")
            ):
                parsed_location = parsed._replace(netloc=parsed_location.netloc)

                return urlunparse(parsed_location)
            elif (
                not parsed.netloc.startswith("www")
            ) & parsed_location.netloc.startswith("www"):
                parsed_location = parsed._replace(netloc=parsed_location.netloc)

                return urlunparse(parsed_location)
        except ValueError:
            return url
    else:
        return url


def check_ipv4_connection() -> str:
    prefix = "IPv4 -> Internet:"
    url = "https://ipv4.icanhazip.com/"

    return f"{prefix} {_check_connection(url)}"


def check_ipv6_connection() -> str:
    prefix = "IPv6 -> Internet:"
    url = "https://ipv6.icanhazip.com/"

    return f"{prefix} {_check_connection(url)}"


def _check_connection(url: str) -> str:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    result = "Connection Failed"

    try:
        headers = {"User-Agent": SERVICE_UA}

        res = requests.get(url, headers=headers, verify=False)

        result = res.text.strip()
    except Exception:
        output.debug_exception()

    return result



from dns import resolver, exception, dnssec
from yawast.shared import output


def get_dnskey(domain):
    records = []

    try:
        answers = resolver.query(domain, "DNSKEY")

        for data in answers:
            flags = format(data.flags, "016b")
            proto = data.protocol
            alg = dnssec.algorithm_to_text(data.algorithm)
            key = data.key

            records.append([flags, proto, alg, key])
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return records



from typing import List, Dict
from requests.models import Response
from urllib.parse import urlparse

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import response_scanner
from yawast.scanner.plugins.result import Result
from yawast.scanner.plugins.http.servers import apache_httpd, php, iis, nginx, python
from yawast.shared import network


def get_header_issues(headers: Dict, raw: str, url: str) -> List[Result]:
    results = []

    if "X-Powered-By" in headers:
        results.append(
            Result(
                f'X-Powered-By Header Present: {headers["X-Powered-By"]} ({url})',
                Vulnerabilities.HTTP_HEADER_X_POWERED_BY,
                url,
                raw,
            )
        )

        # check to see if this is a php version
        results += php.check_version(headers["X-Powered-By"], raw, url)

    if "X-XSS-Protection" in headers:
        # header is present, check the value
        if headers["X-XSS-Protection"] == 0:
            results.append(
                Result(
                    f"X-XSS-Protection Disabled Header Present ({url})",
                    Vulnerabilities.HTTP_HEADER_X_XSS_PROTECTION_DISABLED,
                    url,
                    raw,
                )
            )
    else:
        results.append(
            Result(
                f"X-XSS-Protection Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_X_XSS_PROTECTION_MISSING,
                url,
                raw,
            )
        )

    if "X-Runtime" in headers:
        results.append(
            Result(
                f"X-Runtime Header Present; likely indicates a RoR application ({url})",
                Vulnerabilities.HTTP_HEADER_X_RUNTIME,
                url,
                raw,
            )
        )

    if "X-Backend-Server" in headers:
        results.append(
            Result(
                f'X-Backend-Server Header Present: {headers["X-Backend-Server"]} ({url})',
                Vulnerabilities.HTTP_HEADER_X_BACKEND_SERVER,
                url,
                raw,
            )
        )

    if "Via" in headers:
        results.append(
            Result(
                f'Via Header Present: #{headers["Via"]} ({url})',
                Vulnerabilities.HTTP_HEADER_VIA,
                url,
                raw,
            )
        )

    if "X-Frame-Options" in headers:
        if "allow" in str(headers["X-Frame-Options"]).lower():
            results.append(
                Result(
                    f'X-Frame-Options Header: {headers["X-Frame-Options"]} ({url})',
                    Vulnerabilities.HTTP_HEADER_X_FRAME_OPTIONS_ALLOW,
                    url,
                    raw,
                )
            )
    else:
        results.append(
            Result(
                f"X-Frame-Options Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_X_FRAME_OPTIONS_MISSING,
                url,
                raw,
            )
        )

    if "X-Content-Type-Options" not in headers:
        results.append(
            Result(
                f"X-Content-Type-Options Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_X_CONTENT_TYPE_OPTIONS_MISSING,
                url,
                raw,
            )
        )

    if "Content-Security-Policy" not in headers:
        results.append(
            Result(
                f"Content-Security-Policy Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_CONTENT_SECURITY_POLICY_MISSING,
                url,
                raw,
            )
        )

    if "Referrer-Policy" not in headers:
        results.append(
            Result(
                f"Referrer-Policy Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_REFERRER_POLICY_MISSING,
                url,
                raw,
            )
        )

    if "Feature-Policy" not in headers:
        results.append(
            Result(
                f"Feature-Policy Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_FEATURE_POLICY_MISSING,
                url,
                raw,
            )
        )

    if "Access-Control-Allow-Origin" in headers:
        if headers["Access-Control-Allow-Origin"] == "*":
            results.append(
                Result(
                    f"Access-Control-Allow-Origin: Unrestricted ({url})",
                    Vulnerabilities.HTTP_HEADER_CORS_ACAO_UNRESTRICTED,
                    url,
                    raw,
                )
            )

    if "Strict-Transport-Security" not in headers:
        results.append(
            Result(
                f"Strict-Transport-Security Header Not Present ({url})",
                Vulnerabilities.HTTP_HEADER_HSTS_MISSING,
                url,
                raw,
            )
        )

    if "Server" in headers:
        results += get_server_banner_issues(headers["Server"], raw, url, headers)

    return results


def get_server_banner_issues(
    server: str, raw: str, url: str, headers: Dict
) -> List[Result]:
    results = []

    results += apache_httpd.check_banner(server, raw, url)
    results += nginx.check_banner(server, raw, url)
    results += iis.check_version(server, raw, url, headers)
    results += python.check_banner(server, raw, url)

    return results


def check_propfind(url: str) -> List[Result]:
    results = []

    res = network.http_custom("PROPFIND", url)
    body = res.text

    if res.status_code <= 400 and len(body) > 0:
        if "Content-Type" in res.headers and "text/xml" in res.headers["Content-Type"]:
            results.append(
                Result(
                    "Possible Info Disclosure: PROPFIND Enabled",
                    Vulnerabilities.HTTP_PROPFIND_ENABLED,
                    url,
                    [
                        network.http_build_raw_request(res.request),
                        network.http_build_raw_response(res),
                    ],
                )
            )

    results += response_scanner.check_response(url, res)

    return results


def check_trace(url: str) -> List[Result]:
    results = []

    res = network.http_custom("TRACE", url)
    body = res.text

    if res.status_code == 200 and "TRACE / HTTP/1.1" in body:
        results.append(
            Result(
                "HTTP TRACE Enabled",
                Vulnerabilities.HTTP_TRACE_ENABLED,
                url,
                [
                    network.http_build_raw_request(res.request),
                    network.http_build_raw_response(res),
                ],
            )
        )

    results += response_scanner.check_response(url, res)

    return results


def check_options(url: str) -> List[Result]:
    results = []

    res = network.http_options(url)

    if "Allow" in res.headers:
        results.append(
            Result(
                f"Allow HTTP Verbs (OPTIONS): {res.headers['Allow']}",
                Vulnerabilities.HTTP_OPTIONS_ALLOW,
                url,
                [
                    network.http_build_raw_request(res.request),
                    network.http_build_raw_response(res),
                ],
            )
        )

    if "Public" in res.headers:
        results.append(
            Result(
                f"Public HTTP Verbs (OPTIONS): {res.headers['Allow']}",
                Vulnerabilities.HTTP_OPTIONS_PUBLIC,
                url,
                [
                    network.http_build_raw_request(res.request),
                    network.http_build_raw_response(res),
                ],
            )
        )

    results += response_scanner.check_response(url, res)

    return results


def get_cookie_issues(res: Response, raw: str, url: str) -> List[Result]:
    if "Set-Cookie" in res.headers:
        cookies = res.raw.headers.getlist("Set-Cookie")

        return _get_cookie_issues(cookies, raw, url)
    else:
        return []


def _get_cookie_issues(cookies: List[str], raw: str, url: str) -> List[Result]:
    results = []
    parsed = urlparse(url)

    for cookie in cookies:
        comp = cookie.split(";")

        # get the name
        name = comp[0].split("=")[0]

        # normalize the components
        comp = list(map(str.strip, comp))
        comp = list(map(str.lower, comp))

        # check Secure flag
        if "secure" not in comp:
            if parsed.scheme == "https":
                results.append(
                    Result(
                        f"Cookie Missing Secure Flag: {cookie}",
                        Vulnerabilities.COOKIE_MISSING_SECURE_FLAG,
                        url,
                        [name, raw],
                    )
                )
            else:
                # secure flag over HTTP is invalid
                if "secure" in comp:
                    results.append(
                        Result(
                            f"Cookie Secure Flag Invalid (over HTTP): {cookie}",
                            Vulnerabilities.COOKIE_INVALID_SECURE_FLAG,
                            url,
                            [name, raw],
                        )
                    )

        # check HttpOnly flag
        if "httponly" not in comp:
            results.append(
                Result(
                    f"Cookie Missing HttpOnly Flag: {cookie}",
                    Vulnerabilities.COOKIE_MISSING_HTTPONLY_FLAG,
                    url,
                    [name, raw],
                )
            )

        # check SameSite flag
        if (
            "samesite=lax" not in comp
            and "samesite=strict" not in comp
            and "samesite=none" not in comp
        ):
            results.append(
                Result(
                    f"Cookie Missing SameSite Flag: {cookie}",
                    Vulnerabilities.COOKIE_MISSING_SAMESITE_FLAG,
                    url,
                    [name, raw],
                )
            )

        # check SameSite=None flag
        if "samesite=none" in comp:
            if "secure" in comp:
                results.append(
                    Result(
                        f"Cookie With SameSite=None Flag: {cookie}",
                        Vulnerabilities.COOKIE_WITH_SAMESITE_NONE_FLAG,
                        url,
                        [name, raw],
                    )
                )
            else:
                results.append(
                    Result(
                        f"Cookie SameSite=None Flag Invalid (without Secure flag): {cookie}",
                        Vulnerabilities.COOKIE_INVALID_SAMESITE_NONE_FLAG,
                        url,
                        [name, raw],
                    )
                )

    return results



import os
from multiprocessing import Manager
from multiprocessing.dummy import Pool

import pkg_resources
from dns import resolver, exception

from yawast.shared import output


def find_subdomains(domain, path=None):
    records = []

    # create processing pool
    # given the amount of waiting we do, go for double the CPU count
    pool = Pool(os.cpu_count() * 2)
    mgr = Manager()
    queue = mgr.Queue()

    # read the data in from the data directory
    if path is None:
        file_path = pkg_resources.resource_filename(
            "yawast", "resources/subdomains.txt"
        )
    else:
        file_path = path

    with open(file_path) as file:
        for line in file:
            host = line.strip() + "." + domain

            pool.apply_async(_get_records_for_domain, (host, queue))

    pool.close()
    pool.join()

    while not queue.empty():
        val = queue.get()
        if len(val) > 0:
            records.extend(val)

    return records


def _get_records_for_domain(host, queue):
    records = []

    try:
        answers = resolver.query(host, "CNAME", lifetime=3)

        for data in answers:
            records.append(["CNAME", host, str(data.target)])
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    try:
        answers = resolver.query(host, "A", lifetime=3)

        for data in answers:
            records.append(["A", host, str(data)])
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    try:
        answers = resolver.query(host, "AAAA", lifetime=3)

        for data in answers:
            records.append(["AAAA", host, str(data)])
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    queue.put(records)



from argparse import Namespace

from nassl.ssl_client import OpenSslVersionEnum
from sslyze import server_connectivity_tester
from sslyze.plugins.openssl_cipher_suites_plugin import OPENSSL_TO_RFC_NAMES_MAPPING
from sslyze.utils import (
    ssl_connection_configurator,
    ssl_connection,
    http_response_parser,
)
from validator_collection import checkers

from yawast.scanner.plugins.dns import basic
from yawast.shared import output, utils, network
from yawast.reporting import reporter, issue
from yawast.reporting.enums import Vulnerabilities, Severity


def scan(args: Namespace, url: str, domain: str):
    ips = basic.get_ips(domain)

    for ip in ips:
        conn = None
        count = 0

        try:
            count = 0

            conn_tester = server_connectivity_tester.ServerConnectivityTester(
                hostname=domain, port=utils.get_port(url), ip_address=ip
            )

            output.norm(
                f"TLS Session Request Limit: Checking number of requests accepted using 3DES suites "
                f"(IP: {conn_tester.ip_address}:{conn_tester.port})"
            )

            server_info = conn_tester.perform()
            conn = ssl_connection_configurator.SslConnectionConfigurator.get_connection(
                ssl_version=OpenSslVersionEnum.SSLV23,
                server_info=server_info,
                should_use_legacy_openssl=True,
                openssl_cipher_string="3DES",
                should_ignore_client_auth=True,
                ssl_verify_locations=None,
            )

            conn.connect()

            req = (
                "HEAD / HTTP/1.1\r\n"
                "Host: {host}\r\n"
                "User-Agent: {user_agent}\r\n"
                "Accept: */*\r\n"
                "Connection: keep-alive\r\n\r\n".format(
                    host=domain, user_agent=network.YAWAST_UA
                )
            )

            ossl_name = conn.ssl_client.get_current_cipher_name()
            name = OPENSSL_TO_RFC_NAMES_MAPPING[OpenSslVersionEnum.TLSV1].get(
                ossl_name, ossl_name
            )
            print("       ", end="", flush=True)
            print(f"(using {name})", end="", flush=True)
            for i in range(0, 10000):
                conn.ssl_client.write(req)
                resp = http_response_parser.HttpResponseParser.parse_from_ssl_connection(
                    conn.ssl_client
                )
                count += 1

                if i % 20:
                    print(".", end="", flush=True)

            output.empty()
            reporter.display(
                f"\tTLS Session Request Limit: Connection not terminated after {count} requests; "
                f"possibly vulnerable to SWEET32",
                issue.Issue(Vulnerabilities.TLS_SWEET32, url),
            )

        except ssl_connection.SslHandshakeRejected as error:
            output.debug_exception()

            output.empty()
            output.norm(f"\tServer rejected our connection ({str(error)})")

            output.empty()
        except IOError as error:
            output.debug_exception()

            output.empty()
            if count > 0:
                output.norm(
                    f"\tTLS Session Request Limit: Connection terminated after {count} requests ({str(error)})"
                )
            else:
                output.norm(
                    "\tTLS Session Request Limit: Server does not support 3DES cipher suites"
                )

            output.empty()
        except server_connectivity_tester.ServerConnectivityError as error:
            output.debug_exception()

            output.empty()

            if checkers.is_ipv6(ip):
                output.error(
                    "\tError connecting to IPv6 IP. Please ensure that your system is configured properly."
                )

            output.error(f"\tConnection failed ({str(error)})")
            output.empty()
        finally:
            if conn is not None:
                conn.close()

        output.empty()



import ipaddress
import socket

from publicsuffixlist import PublicSuffixList

from yawast.reporting import reporter, issue
from yawast.reporting.enums import Vulnerabilities, Severity
from yawast.scanner.plugins.dns import basic
from yawast.scanner.plugins.dns import caa
from yawast.scanner.plugins.dns import dnssec
from yawast.scanner.plugins.dns import network_info
from yawast.scanner.plugins.dns import srv
from yawast.scanner.plugins.dns import subdomains
from yawast.shared import utils, output


def scan(args, url, domain):
    # check to see if this is an IP, if so, bail out
    if utils.is_ip(domain):
        return

    output.empty()
    output.norm("DNS Information:")

    # get the root domain, by looking up via the PSL
    psl = PublicSuffixList()
    root_domain = psl.privatesuffix(domain)

    # IP Addresses for the domain we are scanning
    ips = basic.get_ips(domain)
    for ip in ips:
        output.norm("\t%s (%s)" % (ip, basic.get_host(str(ip))))

        addr = ipaddress.ip_address(str(ip))

        if not addr.is_private:
            ni = network_info.network_info(str(ip))
            output.norm("\t\t%s" % ni)

            if addr.version == 4:
                output.norm("\t\thttps://www.shodan.io/host/%s" % ip)
                output.norm("\t\thttps://censys.io/ipv4/%s" % ip)
            else:
                output.norm("\t\thttps://www.shodan.io/host/%s" % str(ip).lower())

        output.empty()

    # TXT records for the domain we are scanning
    txt = basic.get_text(domain)
    for rec in txt:
        output.norm("\tTXT: %s" % rec)

    # TXT records for the root domain
    if root_domain != domain:
        txt = basic.get_text(root_domain)
        for rec in txt:
            output.norm("\tTXT (%s): %s" % (root_domain, rec))

    output.empty()

    # MX records for the domain we are scanning
    mx = basic.get_mx(domain)
    for rec in mx:
        server_ip = socket.gethostbyname(rec[0])
        ni = network_info.network_info(str(server_ip))

        info = "%s (%s) - %s (%s)" % (rec[0], rec[1], server_ip, ni)
        output.norm("\tMX: %s" % info)

    # MX records for the root domain
    if root_domain != domain:
        mx = basic.get_mx(root_domain)
        for rec in mx:
            server_ip = socket.gethostbyname(rec[0])
            ni = network_info.network_info(str(server_ip))

            info = "%s (%s) - %s (%s)" % (rec[0], rec[1], server_ip, ni)
            output.norm("\tMX (%s): %s" % (root_domain, info))

    output.empty()

    # NS records for the root domain
    ns = basic.get_ns(root_domain)
    for rec in ns:
        server_ip = socket.gethostbyname(rec)
        ni = network_info.network_info(str(server_ip))

        info = "%s - %s (%s)" % (rec, server_ip, ni)
        output.norm("\tNS: %s" % info)

    output.empty()

    if args.srv:
        output.norm("Searching for SRV records, this will take a minute...")
        output.empty()

        srv_records = srv.find_srv_records(root_domain)
        for rec in srv_records:
            server_ip = socket.gethostbyname(rec[1])
            ni = network_info.network_info(str(server_ip))

            info = "%s: %s:%s - %s (%s)" % (rec[0], rec[1], rec[2], server_ip, ni)
            output.norm("\tSRV: %s" % info)

        output.empty()

    if args.subdomains:
        output.norm("Searching for sub-domains, this will take a few minutes...")
        output.empty()

        sds = subdomains.find_subdomains(root_domain)
        for rec in sds:
            info = ""

            if rec[0] == "CNAME":
                server_ip = socket.gethostbyname(rec[2])
                ni = network_info.network_info(str(server_ip))

                info = "(CNAME) %s -> %s - %s (%s)" % (rec[1], rec[2], server_ip, ni)
            elif rec[0] == "A":
                ni = network_info.network_info(rec[2])
                info = "(A) %s: %s (%s)" % (rec[1], rec[2], ni)
            elif rec[0] == "AAAA":
                ni = network_info.network_info(rec[2])
                info = "(AAAA) %s: %s (%s)" % (rec[1], rec[2], ni)

            output.norm("\tSubdomain: %s" % info)

        output.empty()

    caa_count = 0
    carec = caa.get_caa(domain)
    for rec in carec:
        curr = rec[0]

        if rec[1] == "CNAME":
            output.norm("\tCAA (%s): CNAME Found: -> %s" % (curr, rec[2]))
        elif rec[1] == "CAA":
            if len(rec[2]) > 0:
                for line in rec[2]:
                    output.norm('\tCAA (%s): "%s"' % (curr, line))
                    caa_count += 1
            else:
                output.norm("\tCAA (%s): No Records Found" % curr)

    output.empty()

    # notify the user if there's an issue
    if caa_count == 0:
        reporter.display(
            "\tCAA: Domain does not have protection from CAA",
            issue.Issue(Vulnerabilities.DNS_CAA_MISSING, url),
        )

    dk = dnssec.get_dnskey(domain)
    if len(dk) > 0:
        for rec in dk:
            output.norm(
                "\tDNSKEY: Algorithm: '%s' - Flags: '%s' - Key Length: %s"
                % (rec[2], rec[0], len(rec[3]) * 8)
            )
    else:
        reporter.display(
            "\tDNSKEY: Domain does not use DNSSEC",
            issue.Issue(Vulnerabilities.DNS_CAA_MISSING, url),
        )

    output.empty()



# From: https://stackoverflow.com/a/36974338
# License: Creative Commons Attribution-Share Alike
# Copyright: Matthew Strax-Haber


def getchar():
    # figure out which function to use once, and store it in _func
    if "_func" not in getchar.__dict__:
        try:
            # for Windows-based systems
            import msvcrt  # If successful, we are on Windows

            getchar._func = msvcrt.getch

        except ImportError:
            # for POSIX-based systems (with termios & tty support)
            import tty, sys, termios, fcntl, os

            def _ttyread():
                fd = sys.stdin.fileno()

                oldterm = termios.tcgetattr(fd)
                newattr = termios.tcgetattr(fd)
                newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
                termios.tcsetattr(fd, termios.TCSANOW, newattr)

                oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
                fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)

                try:
                    while 1:
                        try:
                            answer = sys.stdin.read(1)

                            if answer == "\x03":
                                raise KeyboardInterrupt

                            break
                        except IOError:
                            pass
                finally:
                    termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
                    fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)

                return answer

            getchar._func = _ttyread

    return getchar._func()



from dns import resolver, exception, reversename
from yawast.shared import output


def get_ips(domain: str):
    ips = []

    try:
        answers_v4 = resolver.query(domain, "A")

        for data in answers_v4:
            ips.append(str(data))
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    try:
        answers_v6 = resolver.query(domain, "AAAA")
        for data in answers_v6:
            ips.append(str(data))
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return ips


def get_text(domain):
    records = []

    try:
        answers = resolver.query(domain, "TXT")

        for data in answers:
            records.append(str(data))
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return records


def get_mx(domain):
    records = []

    try:
        answers = resolver.query(domain, "MX")

        for data in answers:
            records.append([str(data.exchange), str(data.preference)])
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return records


def get_ns(domain):
    records = []

    try:
        answers = resolver.query(domain, "NS")

        for data in answers:
            records.append(str(data))
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return records


def get_host(ip):
    name = "N/A"

    try:
        rev_name = reversename.from_address(str(ip))
        name = str(resolver.query(rev_name, "PTR", lifetime=3)[0])[:-1]
    except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
        pass
    except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
        output.debug_exception()
        pass

    return name



import re
from typing import Union, List, cast

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result
from requests import Response

from yawast.shared import network, output

_data = []
_reports = []


class _MatchRule:
    def __init__(self, data: str):
        fields = data.split("\t")

        # clean up regex, to eliminate issues from using Java-flavored regex
        pattern = fields[0].replace("}+", "}").replace("++", "+")
        self.pattern = re.compile(pattern)

        self.match_group = fields[1]
        self.type = fields[2]
        self.confidence = fields[4]


def check_response(
    url: str, res: Response, body: Union[str, None] = None
) -> List[Result]:
    global _data, _reports

    # make sure we actually have something
    if res is None:
        return []

    if _data is None or len(_data) == 0:
        _get_data()

    results = []

    if body is None:
        body = res.text

    for rule in _data:
        rule = cast(_MatchRule, rule)

        mtch = re.search(rule.pattern, body)

        if mtch:
            val = mtch.group(rule.match_group)

            err_start = body.find(val)

            # get the error, plus 25 characters on each side
            err = body[err_start - 25 : err_start + len(val) + 25]
            msg = f"Found {rule.type} error message (confidence: {rule.confidence}) on {url}: {err}"

            if msg not in _reports:
                results.append(
                    Result(
                        msg,
                        Vulnerabilities.HTTP_ERROR_MESSAGE,
                        url,
                        [
                            network.http_build_raw_request(res.request),
                            network.http_build_raw_response(res),
                        ],
                    )
                )

                _reports.append(msg)
            else:
                output.debug(f"Ignored duplicate error message: {msg}")

    return results


def reset():
    global _reports

    _reports = []


def _get_data() -> None:
    global _data
    data_url = "https://raw.githubusercontent.com/augustd/burp-suite-error-message-checks/master/src/main/resources/burp/match-rules.tab"

    try:
        raw = network.http_get(data_url).text

        for line in raw.splitlines():
            _data.append(_MatchRule(line))

    except Exception as error:
        output.debug(f"Failed to get version data: {error}")
        output.debug_exception()



from typing import List, cast
from packaging import version

from yawast.scanner.plugins.http import version_checker
from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result


def check_version(banner: str, raw: str, url: str) -> List[Result]:
    if not banner.startswith("PHP/"):
        return []

    results = []

    # we've got a PHP version
    results.append(
        Result(
            f"PHP Version Exposed: {banner}",
            Vulnerabilities.HTTP_PHP_VERSION_EXPOSED,
            url,
            raw,
        )
    )

    # parse the version, and get the latest version - see if the server is up to date
    ver = cast(version.Version, version.parse(banner.split("/")[1]))
    curr_version = version_checker.get_latest_version("php", ver)

    if curr_version is not None and curr_version > ver:
        results.append(
            Result(
                f"PHP Outdated: {ver} - Current: {curr_version}",
                Vulnerabilities.SERVER_PHP_OUTDATED,
                url,
                raw,
            )
        )

    return results



import re
import sys
from urllib.parse import urlparse
from urllib.parse import urlunparse

from validator_collection import checkers

from yawast.shared import output
from yawast.shared.exec_timer import ExecutionTimer


def is_url(url):
    try:
        url = extract_url(url)

        if checkers.is_url(url):
            parsed = urlparse(url)

            if parsed.scheme == "http" or parsed.scheme == "https":
                # make sure the data we have is at least valid-ish
                return all(
                    [
                        parsed.scheme,
                        parsed.netloc,
                        (parsed.port is None or parsed.port > 0),
                    ]
                )
            else:
                return False
        else:
            return False
    except ValueError:
        return False


def is_ip(val):
    # strip any wrapping for an IPv6 used in a URL
    val = str(val).lstrip("[").rstrip("]")

    return checkers.is_ip_address(val)


def get_domain(val):
    val = str(val)
    val = urlparse(extract_url(val)).netloc

    # strip any credentials
    if "@" in val:
        val = val.rpartition("@")[-1]

    # strip any port number
    if ":" in val:
        # we check for an ending ']' because of IPv6
        if not val.endswith("]"):
            val = val.rpartition(":")[0]

    return val


def is_printable_str(b: bytes) -> bool:
    decoders = ["utf_8", "latin_1", "cp1251"]
    printable = False
    good_decoder = None

    with ExecutionTimer() as timer:
        for decoder in decoders:
            s = b.decode(decoder, "backslashreplace")

            if not any(
                repr(ch).startswith("'\\x") or repr(ch).startswith("'\\u") for ch in s
            ):
                printable = True
                good_decoder = decoder

                break

    if good_decoder is not None:
        output.debug(f"Decoded string as {good_decoder} in {timer.to_ms()}ms")

    return printable


def get_port(url: str) -> int:
    parsed = urlparse(url)
    url = parsed.netloc

    if parsed.port is not None:
        return int(parsed.port)

    # strip any credentials
    if "@" in url:
        url = url.rpartition("@")[-1]

    # get any port number
    if ":" in url:
        # we check for an ending ']' because of IPv6
        if not url.endswith("]"):
            return int(url.rpartition(":")[1])

    if parsed.scheme == "https":
        return 443
    elif parsed.scheme == "http:":
        return 80


def extract_url(url):
    # check for extra slashes in the scheme
    if re.match(r"^[a-z]{2,8}:///+", url, re.IGNORECASE):
        url = re.sub(r":///+", "://", url, 1)

    # check to see if we already have something that looks like a valid scheme
    # we'll only process the cleanup if we don't match
    if not re.match(r"^[a-z]{2,8}://", url, re.IGNORECASE):
        # fix a missing colon
        if url.lower().startswith("http//") or url.lower().startswith("https//"):
            url = url.replace("//", "://", 1)

        # fix URLs that a missing a slash after the scheme
        if re.match(r"^http[s]?:/[^/]", url, re.IGNORECASE):
            url = url.replace(":/", "://", 1)

        # this might be buggy - actually, I know it is...
        # if the URL is malformed, this can lead to some very wrong things
        if not (
            url.lower().startswith("http://") or url.lower().startswith("https://")
        ):
            url = "http://" + url

    # parse the URL so that we can get into the more detailed cleanup
    parsed = urlparse(url)

    # force name to lower, if it isn't
    if parsed.netloc != parsed.netloc.lower():
        parsed = parsed._replace(netloc=parsed.netloc.lower())

    # make sure we have something set for the path
    if parsed.path == "":
        parsed = parsed._replace(path="/")

    # make sure that we are looking at root (most common) or a folder, not a file
    if not parsed.path.endswith("/"):
        # this isn't a great solution, but for now...
        # strip everything after the last slash
        new_path = parsed.path.rsplit("/", 1)[0] + "/"
        parsed = parsed._replace(path=new_path)

    # remove any query strings. not something we can work with
    if not parsed.query == "":
        parsed = parsed._replace(query="")

    # remove any fragment strings. not something we can work with
    if not parsed.fragment == "":
        parsed = parsed._replace(fragment="")

    # remove any parameters. not something we can work with
    if not parsed.params == "":
        parsed = parsed._replace(params="")

    return urlunparse(parsed)


def exit_message(message):
    print(message, file=sys.stderr)
    sys.exit(-1)



from typing import List, cast, Dict
from packaging import version

from yawast.scanner.plugins.http import version_checker, response_scanner
from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result
from yawast.shared import network, output


def check_all(url: str) -> List[Result]:
    results = []

    results += check_asp_net_debug(url)

    return results


def check_version(banner: str, raw: str, url: str, headers: Dict) -> List[Result]:
    if not banner.startswith("Microsoft-IIS/"):
        return []

    results = []

    # we've got an IIS version
    results.append(
        Result(
            f"IIS Version Exposed: {banner}",
            Vulnerabilities.HTTP_BANNER_IIS_VERSION,
            url,
            raw,
        )
    )

    # parse the version, and get the latest version - see if the server is up to date
    ver = cast(version.Version, version.parse(banner.split("/")[1]))
    curr_version = version_checker.get_latest_version("iis", ver)

    if curr_version is not None and curr_version > ver:
        results.append(
            Result(
                f"IIS Outdated: {ver} - Current: {curr_version}",
                Vulnerabilities.SERVER_IIS_OUTDATED,
                url,
                raw,
            )
        )

    # IIS servers may expose a couple other versions, related to ASP.NET, check for those
    if "X-AspNetMvc-Version" in headers:
        results.append(
            Result(
                f'ASP.NET MVC Version Exposed: {headers["X-AspNetMvc-Version"]}',
                Vulnerabilities.HTTP_HEADER_X_ASPNETMVC_VERSION,
                url,
                raw,
            )
        )

        ver = cast(version.Version, version.parse(headers["X-AspNetMvc-Version"]))
        curr_version = version_checker.get_latest_version("aspnet_mvc", ver)

        if curr_version is not None and curr_version > ver:
            results.append(
                Result(
                    f"ASP.NET MVC Outdated: {ver} - Current: {curr_version}",
                    Vulnerabilities.SERVER_ASPNETMVC_OUTDATED,
                    url,
                    raw,
                )
            )

    if "X-AspNet-Version" in headers:
        results.append(
            Result(
                f'ASP.NET  Version Exposed: {headers["X-AspNet-Version"]}',
                Vulnerabilities.HTTP_HEADER_X_ASPNET_VERSION,
                url,
                raw,
            )
        )

        ver = cast(version.Version, version.parse(headers["X-AspNet-Version"]))
        curr_version = version_checker.get_latest_version("aspnet", ver)

        if curr_version is not None and curr_version > ver:
            results.append(
                Result(
                    f"ASP.NET Outdated: {ver} - Current: {curr_version}",
                    Vulnerabilities.SERVER_ASPNET_OUTDATED,
                    url,
                    raw,
                )
            )

    return results


def check_asp_net_debug(url: str) -> List[Result]:
    results = []

    res = network.http_custom(
        "DEBUG", url, additional_headers={"Command": "stop-debug", "Accept": "*/*"}
    )

    if res.status_code == 200:
        # we've got a hit, but could be a false positive
        # try this again, with a different verb
        xres = network.http_custom(
            "XDEBUG", url, additional_headers={"Command": "stop-debug", "Accept": "*/*"}
        )

        # if we get a 200 when using an invalid verb, it's a false positive
        # if we get something else, then the DEBUG actually did something
        if xres.status_code != 200:
            results.append(
                Result(
                    "ASP.NET Debugging Enabled",
                    Vulnerabilities.SERVER_ASPNET_DEBUG_ENABLED,
                    url,
                    [
                        network.http_build_raw_request(res.request),
                        network.http_build_raw_response(res),
                    ],
                )
            )
        else:
            output.debug("Server responds to invalid HTTP verbs with status 200")

    results += response_scanner.check_response(url, res)

    return results



import re
from typing import Tuple, Union, List, cast
from urllib.parse import urljoin

from packaging import version
from requests import Response

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import version_checker
from yawast.scanner.plugins.result import Result
from yawast.shared import network


def identify(url: str) -> Tuple[Union[str, None], List[Result]]:
    results = []

    # find WordPress
    res, path = _identify_by_path(url, "")

    if path is None:
        res, path = _identify_by_path(url, "blog/")

    # check to see if we have a valid hit
    if path is not None:
        # we have a WordPress install, let's see if we can get a version
        body = res.text

        ver = "Unknown"
        # this works for modern versions
        m = re.search(r"login.min.css\?ver=\d+\.\d+\.?\d*", body)
        if m:
            ver = m.group(0).split("=")[1]
        else:
            # the current method doesn't work, fall back to an older method
            m = re.search(r"load-styles.php\?[\w,;=&%]+;ver=\d+\.\d+\.?\d*", body)
            if m:
                ver = m.group(0).split("=")[-1]

        # report that we found WordPress
        results.append(
            Result(
                f"Found WordPress v{ver} at {path}",
                Vulnerabilities.APP_WORDPRESS_VERSION,
                path,
                [ver, network.http_build_raw_response(res)],
            )
        )

        # is this a current version?
        ver = cast(version.Version, version.parse(ver))
        curr_version = version_checker.get_latest_version("wordpress", ver)

        if curr_version is not None and curr_version > ver:
            results.append(
                Result(
                    f"WordPress Outdated: {ver} - Current: {curr_version}",
                    Vulnerabilities.APP_WORDPRESS_OUTDATED,
                    path,
                    [str(ver), network.http_build_raw_response(res)],
                )
            )

        return path, results
    else:
        return None, []


def check_json_user_enum(url: str) -> List[Result]:
    results = []
    target = urljoin(url, "wp-json/wp/v2/users")

    res = network.http_get(target, False)
    body = res.text

    if res.status_code < 300 and "slug" in body:
        data = res.json()

        # log the enum finding
        results.append(
            Result(
                f"WordPress WP-JSON User Enumeration at {target}",
                Vulnerabilities.APP_WORDPRESS_USER_ENUM_API,
                target,
                [
                    network.http_build_raw_request(res.request),
                    network.http_build_raw_response(res),
                ],
            )
        )

        # log the individual users
        for user in data:
            results.append(
                Result(
                    f"ID: {user['id']}\tUser Slug: '{user['slug']}'\t\tUser Name: '{user['name']}'",
                    Vulnerabilities.APP_WORDPRESS_USER_FOUND,
                    target,
                    [
                        network.http_build_raw_request(res.request),
                        network.http_build_raw_response(res),
                        user["id"],
                        user["slug"],
                        user["name"],
                    ],
                )
            )

    return results


def _identify_by_path(url: str, path: str) -> Tuple[Response, Union[str, None]]:
    target = urljoin(url, f"{path}wp-login.php")

    res = network.http_get(target, False)
    body = res.text

    if res.status_code == 200 and "Powered by WordPress" in body:
        return res, urljoin(url, path)
    else:
        return res, None



import json
from typing import List, Dict, Tuple
from urllib.parse import urljoin, urlparse

from bs4 import BeautifulSoup

from yawast.external import retirejs
from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result
from yawast.shared import output, network, utils

_data = {}
_checked = []
_reports = []


def get_results(soup: BeautifulSoup, url: str, raw: str) -> List[Result]:
    global _reports

    results = []

    parsed = urlparse(url)
    domain = utils.get_domain(parsed.netloc)

    issues, r = _get_retirejs_results(soup, url, domain)
    results += r
    for js_url, issue in issues:
        comp = issue["component"]
        ver = issue["version"]

        if "vulnerabilities" in issue:
            for vuln in issue["vulnerabilities"]:
                info = (
                    f'Vulnerable JavaScript: {comp}-{ver} ({js_url}): Severity: {vuln["severity"]} - '
                    f'Info: {" ".join(vuln["info"])}'
                )

                # make sure we haven't reported this issue before
                if info not in _reports:
                    _reports.append(info)

                    results.append(
                        Result(
                            info,
                            Vulnerabilities.JS_VULNERABLE_VERSION,
                            js_url,
                            [url, raw],
                        )
                    )

    return results


def reset():
    global _checked, _reports

    _checked = []
    _reports = []


def _get_retirejs_results(
    soup: BeautifulSoup, url: str, domain: str
) -> Tuple[List[Tuple[str, Dict]], List[Result]]:
    global _data, _checked
    issues = []
    results = []

    if _data is not None and len(_data) == 0:
        _get_data()

    if _data is not None:
        # get all the JS files
        files = [i.get("src") for i in soup.find_all("script") if i.get("src")]

        for file in files:
            # fix relative URLs
            if str(file).startswith("//"):
                file = f"https:{file}"
            if str(file).startswith("/") or (not str(file).startswith("http")):
                file = urljoin(url, file)

            if file not in _checked:
                findings = retirejs.scan_endpoint(file, _data)

                _checked.append(file)

                if domain not in file:
                    # external JS file
                    results.append(
                        Result(
                            f"External JavaScript File: {file}",
                            Vulnerabilities.JS_EXTERNAL_FILE,
                            file,
                            [url, file],
                        )
                    )

                for find in findings:
                    issues.append((file, find))

        return issues, results
    else:
        # this means we couldn't get the data, so bail
        return [], []


def _get_data() -> None:
    global _data
    data = None
    data_url = "https://raw.githubusercontent.com/RetireJS/retire.js/master/repository/jsrepository.json"

    try:
        raw = network.http_get(data_url).content
        raw = raw.decode("utf-8").replace("§§version§§", "[0-9][0-9.a-z_\\\\-]+")

        data = json.loads(raw)

    except Exception as error:
        output.debug(f"Failed to get version data: {error}")
        output.debug_exception()

    _data = data



import uuid
from typing import Union, List, cast

from yawast.reporting.enums import Vulnerabilities, VulnerabilityInfo
from yawast.scanner.plugins.result import Result
from yawast.shared import output


class Issue(dict):
    def __init__(
        self,
        vuln: Vulnerabilities,
        url: str,
        evidence: Union[str, List[str], None] = None,
    ):
        val = cast(VulnerabilityInfo, vuln.value)

        self.vulnerability = vuln
        self.severity = val.severity
        self.url = url
        self.evidence = evidence
        self.id = uuid.uuid4().hex

        dict.__init__(self, id=self.id, url=self.url, evidence=evidence)

        output.debug(
            f"Issue Created: {self.id} - {self.vulnerability.name} - {self.url}"
        )

    def __repr__(self):
        return f"Result: {self.id} - {self.vulnerability.name} - {self.url}"

    @classmethod
    def from_result(cls, result: Result):
        iss = cls(result.vulnerability, result.url, result.evidence)

        output.debug(f"Issue {iss.id} created from result {result.id}")

        return iss



import locale
import platform
import signal
import ssl
import sys
import threading
import time
from datetime import datetime
from multiprocessing import current_process, active_children

import psutil
from psutil import Process

from yawast import command_line
from yawast._version import get_version
from yawast.external.get_char import getchar
from yawast.external.memory_size import Size
from yawast.reporting import reporter
from yawast.shared import output, network

_start_time = datetime.now()
_monitor = None


def main():
    global _start_time, _monitor

    signal.signal(signal.SIGINT, signal_handler)

    parser = command_line.build_parser()
    args, urls = parser.parse_known_args()

    network.init(args.proxy, args.cookie)

    # setup the output system
    output.setup(args.debug, args.nocolors)

    # if we made it this far, it means that the parsing worked.
    command_line.process_urls(urls)

    # we are good to keep going
    print_header()

    if args.output is not None:
        reporter.init(args.output)
        _set_basic_info()

        print(f"Saving output to '{reporter.get_output_file()}'")
        print()

    try:
        with _KeyMonitor():
            with _ProcessMonitor() as pm:
                _monitor = pm

                args.func(args, urls)
    except KeyboardInterrupt:
        output.empty()
        output.error("Scan cancelled by user.")
    finally:
        _shutdown()


def print_header():
    start_time = time.strftime("%Y-%m-%d %H:%M:%S %Z (%z)", time.localtime())

    vm = psutil.virtual_memory()
    mem_total = "{0:cM}".format(Size(vm.total))
    mem_avail = "{0:cM}".format(Size(vm.available))

    cpu_freq = psutil.cpu_freq()
    cpu_max = int(cpu_freq.max)
    if cpu_max == 0:
        # in this case, we don't have a real max, so go with current
        cpu_max = int(cpu_freq.current)

    print(r"__   _____  _    _  ___   _____ _____ ")
    print(r"\ \ / / _ \| |  | |/ _ \ /  ___|_   _|")
    print(r" \ V / /_\ \ |  | / /_\ \\ `--.  | |  ")
    print(r"  \ /|  _  | |/\| |  _  | `--. \ | |  ")
    print(r"  | || | | \  /\  / | | |/\__/ / | |  ")
    print(r"  \_/\_| |_/\/  \/\_| |_/\____/  \_/  ")
    print("")
    print(
        f"YAWAST v{get_version()} - The YAWAST Antecedent Web Application Security Toolkit"
    )
    print(" Copyright (c) 2013-2019 Adam Caudill <adam@adamcaudill.com>")
    print(" Support & Documentation: https://github.com/adamcaudill/yawast")
    print(
        f" Python {''.join(sys.version.splitlines())} ({platform.python_implementation()})"
    )
    print(f" {ssl.OPENSSL_VERSION}")
    print(f" Platform: {platform.platform()} ({_get_locale()})")
    print(
        f" CPU(s): {psutil.cpu_count()}@{cpu_max}MHz - RAM: {mem_total} ({mem_avail} Available)"
    )
    print(f" Started at {start_time}")
    print("")

    print("Connection Status:")
    print(f" {network.check_ipv4_connection()}")
    print(f" {network.check_ipv6_connection()}")
    print("")


def signal_handler(sig, frame):
    if sig == signal.SIGINT:
        # check to see if we are a worker, or the main process
        if current_process().name == "MainProcess":
            print("Scan cancelled by user.")
            _shutdown()

        try:
            active_children()
        except:
            # we don't care if this fails
            pass

        sys.exit(1)


def _shutdown():
    global _start_time, _monitor

    elapsed = datetime.now() - _start_time
    mem_res = "{0:cM}".format(Size(_monitor.peak_mem_res))

    output.empty()

    if reporter.get_output_file() != "":
        reporter.save_output()

    output.norm(f"Completed (Elapsed: {str(elapsed)} - Peak Memory: {mem_res})")


def _get_locale() -> str:
    # get the locale
    try:
        locale.setlocale(locale.LC_ALL, "")
        lcl = locale.getdefaultlocale()

    except Exception as error:
        print(
            f"Unable to get Locale: {str(error)} - attempting to force locale to en_US.utf8"
        )

        try:
            if platform.system() == "Darwin":
                locale.setlocale(locale.LC_ALL, "EN_US")
            else:
                locale.setlocale(locale.LC_ALL, "en_US.utf8")

            lcl = locale.getdefaultlocale()
        except Exception as err:
            print(f"Unable to set locale: {str(err)}")

            lcl = None

    if lcl is not None:
        loc = f"{lcl[0]}.{lcl[1]}"
    else:
        loc = "(Unknown locale)"

    return loc


def _set_basic_info():
    reporter.register_info("start_time", int(time.time()))
    reporter.register_info("yawast_version", get_version())
    reporter.register_info(
        "python_version",
        f"{''.join(sys.version.splitlines())} ({platform.python_implementation()})",
    )
    reporter.register_info("openssl_version", ssl.OPENSSL_VERSION)
    reporter.register_info("platform", platform.platform())
    reporter.register_info("options", str(sys.argv))
    reporter.register_info("encoding", _get_locale())


class _KeyMonitor:
    busy = False

    def wait_task(self):
        if sys.stdout.isatty():
            while self.busy:
                try:
                    key = getchar()

                    if key != "":
                        output.debug(f"Received from keyboard: {key}")

                        if key == "d":
                            output.toggle_debug()

                    time.sleep(0.1)
                except Exception:
                    output.debug_exception()

                    self.busy = False

                    pass
        else:
            # if this isn't a TTY, no point in doing any of this
            self.busy = False

    def __enter__(self):
        self.busy = True
        threading.Thread(target=self.wait_task).start()

    def __exit__(self, exception, value, tb):
        self.busy = False

        if exception is not None:
            return False


class _ProcessMonitor:
    WARNING_THRESHOLD = 100 * 1024 * 1024

    busy = False

    def __init__(self):
        self.process = Process()
        self.peak_mem_res = 0
        self.low_mem_warning = False

    def monitor_task(self):
        if sys.stdout.isatty():

            while self.busy:
                try:
                    # only print the data out every 10 seconds
                    if datetime.now().second / 10 == 0:
                        info = self._get_info()

                        output.debug(info)
                    else:
                        # call get_mem so that we record peak more accurately
                        self._get_mem()

                    time.sleep(1)
                except Exception:
                    output.debug_exception()

                    self.busy = False

                    pass
        else:
            # if this isn't a TTY, no point in doing any of this
            self.busy = False

    def _get_info(self) -> str:
        from yawast.external.memory_size import Size

        # prime the call to cpu_percent, as the first call doesn't return useful data
        self.process.cpu_percent(interval=1)

        # use oneshot() to cache the data, so we minimize hits
        with self.process.oneshot():
            pct = self.process.cpu_percent()

            times = self.process.cpu_times()
            mem = self._get_mem()
            mem_res = "{0:cM}".format(Size(mem.rss))
            mem_virt = "{0:cM}".format(Size(mem.vms))

            thr = self.process.num_threads()

            vm = psutil.virtual_memory()
            mem_total = "{0:cM}".format(Size(vm.total))
            mem_avail_bytes = vm.available
            mem_avail = "{0:cM}".format(Size(vm.available))

            if mem_avail_bytes < self.WARNING_THRESHOLD and not self.low_mem_warning:
                self.low_mem_warning = True

                output.error(f"Low RAM Available: {mem_avail}")

            cons = -1
            try:
                cons = len(self.process.connections(kind="inet"))
            except Exception:
                # we don't care if this fails
                output.debug_exception()

            cpu_freq = psutil.cpu_freq()

        info = (
            f"Process Stats: CPU: {pct}% - Sys: {times.system} - "
            f"User: {times.user} - Res: {mem_res} - Virt: {mem_virt} - "
            f"Available: {mem_avail}/{mem_total} - Threads: {thr} - "
            f"Connections: {cons} - CPU Freq: "
            f"{int(cpu_freq.current)}MHz/{int(cpu_freq.max)}MHz"
        )

        return info

    def _get_mem(self):
        mem = self.process.memory_info()

        if mem.rss > self.peak_mem_res:
            self.peak_mem_res = mem.rss

        return mem

    def __enter__(self):
        self.busy = True
        threading.Thread(target=self.monitor_task).start()

        return self

    def __exit__(self, exception, value, tb):
        self.busy = False

        if exception is not None:
            return False



from yawast.shared import network, utils, output

API_SERVER = "https://api.ssllabs.com"


def get_info_message():
    path = "/api/v3/info"
    messages = []

    try:
        body, code = network.http_json(API_SERVER + path)

        if len(body["messages"]) > 0:
            for msg in body["messages"]:
                messages.append(msg)
    except Exception:
        output.debug_exception()
        raise

    return messages


def start_scan(domain):
    resp = _analyze(domain, True)
    status = resp["status"]

    output.debug(f"Started SSL Labs scan: {resp}")

    return status, resp


def check_scan(domain):
    resp = _analyze(domain)
    status = resp["status"]

    if status != "READY":
        output.debug(f"SSL Labs status: {resp}")

    return status, resp


def _analyze(domain, new=False):
    new_path = "host={target}&publish=off&startNew=on&all=done&ignoreMismatch=on".format(
        target=domain
    )
    status_path = "host={target}&publish=off&all=done&ignoreMismatch=on".format(
        target=domain
    )

    if new:
        path = new_path
    else:
        path = status_path

    try:
        body, code = network.http_json(API_SERVER + "/api/v3/analyze?" + path)
    except Exception:
        output.debug_exception()
        raise

    # check for error messages
    if body.get("errors") is not None:
        raise ValueError(
            "SSL Labs returned the following error(s): {errors}".format(
                errors=str(body["errors"])
            )
        )

    # next up, check to see what error code we have
    if code != 200:
        # if we got anything but 200, it's a problem
        raise ValueError("SSL Labs returned error code: {code}".format(code=code))

    return body



import pkg_resources
from dns import resolver, exception
from yawast.shared import output


def find_srv_records(domain, path=None):
    records = []

    # read the data in from the data directory
    if path is None:
        file_path = pkg_resources.resource_filename("yawast", "resources/srv.txt")
    else:
        file_path = path

    with open(file_path) as file:
        for line in file:
            host = line.strip() + "." + domain

            try:
                answers = resolver.query(host, "SRV", lifetime=3)

                for data in answers:
                    target = data.target.to_text()
                    port = str(data.port)

                    records.append([host, target, port])
            except (resolver.NoAnswer, resolver.NXDOMAIN, exception.Timeout):
                pass
            except (resolver.NoNameservers, resolver.NotAbsolute, resolver.NoRootSOA):
                output.debug_exception()
                pass

    return records



# From: https://github.com/ActiveState/code/blob/master/recipes/Python/578323_Humreadable_filememory_sizes/recipe-578323.py
# License: MIT
# Copyright: ActiveState

import math


class Size(int):
    """ define a size class to allow custom formatting
        format specifiers supported :
            em : formats the size as bits in IEC format i.e. 1024 bits (128 bytes) = 1Kib
            eM : formats the size as Bytes in IEC format i.e. 1024 bytes = 1KiB
            sm : formats the size as bits in SI format i.e. 1000 bits = 1kb
            sM : formats the size as bytes in SI format i.e. 1000 bytes = 1KB
            cm : format the size as bit in the common format i.e. 1024 bits (128 bytes) = 1Kb
            cM : format the size as bytes in the common format i.e. 1024 bytes = 1KB
    """

    def __format__(self, fmt):
        # is it an empty format or not a special format for the size class
        if fmt == "" or fmt[-2:].lower() not in ["em", "sm", "cm"]:
            if fmt[-1].lower() in ["b", "c", "d", "o", "x", "n", "e", "f", "g", "%"]:
                # Numeric format.
                return int(self).__format__(fmt)
            else:
                return str(self).__format__(fmt)

        # work out the scale, suffix and base
        factor, suffix = (8, "b") if fmt[-1] in str(self).lower() else (1, "B")
        base = 1024 if fmt[-2] in ["e", "c"] else 1000

        # Add the i for the IEC format
        suffix = "i" + suffix if fmt[-2] == "e" else suffix

        mult = ["", "K", "M", "G", "T", "P"]

        val = float(self) * factor
        i = 0 if val < 1 else int(math.log(val, base)) + 1
        v = val / math.pow(base, i)
        v, i = (v, i) if v > 0.5 else (v * base, i - 1)

        # Identify if there is a width and extract it
        width = "" if fmt.find(".") == -1 else fmt[: fmt.index(".")]
        precis = fmt[:-2] if width == "" else fmt[fmt.index(".") : -2]

        if precis == "":
            precis = ".2"

        # do the precision bit first, so width/alignment works with the suffix
        t = ("{0:{1}f}" + mult[i] + suffix).format(v, precis)

        return "{0:{1}}".format(t, width) if width != "" else t



from typing import List, Tuple
from urllib.parse import urljoin

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import response_scanner
from yawast.scanner.plugins.result import Result
from yawast.shared import network


def check_special_files(url: str) -> Tuple[List[str], List[Result]]:
    targets = [
        "crossdomain.xml",
        "clientaccesspolicy.xml",
        "sitemap.xml",
        "WS_FTP.LOG",
        "ws_ftp.log",
        "Trace.axd",
        "elmah.axd",
        "readme.html",
        "RELEASE-NOTES.txt",
        "docs/RELEASE-NOTES.txt",
        "CHANGELOG.txt",
        "core/CHANGELOG.txt",
        "license.txt",
    ]

    return _check_url(url, targets)


def check_special_paths(url: str) -> Tuple[List[str], List[Result]]:
    targets = [".git/", ".hg/", ".svn/", ".bzr/", ".cvs/"]

    return _check_url(url, targets)


def _check_url(url: str, targets: List[str]) -> Tuple[List[str], List[Result]]:
    files = []
    results = []

    for target in targets:
        target_url = urljoin(url, target)

        res = network.http_get(target_url, False)

        results += response_scanner.check_response(target_url, res)

        if res.status_code < 300:
            files.append(target_url)
            results.append(
                Result(
                    f"File found: {target_url}",
                    Vulnerabilities.SERVER_SPECIAL_FILE_EXPOSED,
                    target_url,
                    res.text,
                )
            )

    return files, results



from typing import List

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result


def check_banner(banner: str, raw: str, url: str) -> List[Result]:
    if not banner.startswith("Python/"):
        return []

    results = []

    # we've got a version
    results.append(
        Result(
            f"Python Version Exposed: {banner}",
            Vulnerabilities.HTTP_BANNER_PYTHON_VERSION,
            url,
            raw,
        )
    )

    return results



from argparse import Namespace
from typing import List

from yawast.external.spinner import Spinner
from yawast.reporting import reporter
from yawast.reporting.enums import Vulnerabilities
from yawast.reporting.issue import Issue
from yawast.scanner.plugins.http import (
    http_basic,
    waf,
    spider,
    retirejs,
    special_files,
    file_search,
    error_checker,
)
from yawast.scanner.plugins.http.applications import wordpress
from yawast.scanner.plugins.http.servers import apache_httpd, apache_tomcat, nginx, iis
from yawast.shared import network, output


def scan(args: Namespace, url: str, domain: str):
    output.empty()
    output.norm("HEAD:")
    head = network.http_head(url)

    raw = network.http_build_raw_response(head)
    for line in raw.splitlines():
        output.norm(f"\t{line}")

    output.empty()

    res = http_basic.get_header_issues(head.headers, raw, url)
    if len(res) > 0:
        output.norm("Header Issues:")

        reporter.display_results(res, "\t")
        output.empty()

    res = http_basic.get_cookie_issues(head, raw, url)
    if len(res) > 0:
        output.norm("Cookie Issues:")

        reporter.display_results(res, "\t")
        output.empty()

    # check for WAF signatures
    res = waf.get_waf(head.headers, raw, url)
    if len(res) > 0:
        output.norm("WAF Detection:")

        reporter.display_results(res, "\t")
        output.empty()

    output.norm("Performing vulnerability scan (this will take a while)...")

    links = []
    with Spinner():
        try:
            links, res = spider.spider(url)
        except Exception as error:
            output.debug_exception()
            output.error(f"Error running scan: {str(error)}")

    output.norm(f"Identified {len(links) + 1} pages.")
    output.empty()

    if len(res) > 0:
        output.norm("Issues Detected:")

        reporter.display_results(res, "\t")
        output.empty()

    # get files, and add those to the link list
    links += _file_search(args, url, links)

    res = apache_httpd.check_all(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = apache_tomcat.check_all(url, links)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = nginx.check_all(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = iis.check_all(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = http_basic.check_propfind(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = http_basic.check_trace(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    res = http_basic.check_options(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    wp_path, res = wordpress.identify(url)
    if len(res) > 0:
        reporter.display_results(res, "\t")

    if wp_path is not None:
        res = wordpress.check_json_user_enum(wp_path)
        if len(res) > 0:
            reporter.display_results(res, "\t")


def reset():
    retirejs.reset()
    file_search.reset()
    error_checker.reset()


def _file_search(args: Namespace, url: str, orig_links: List[str]) -> List[str]:
    new_files = []
    file_good, path_good = network.check_404_response(url)

    if not file_good:
        reporter.display(
            "Web server does not respond properly to file 404 errors.",
            Issue(Vulnerabilities.SERVER_INVALID_404_FILE, url),
        )
    if not path_good:
        reporter.display(
            "Web server does not respond properly to path 404 errors.",
            Issue(Vulnerabilities.SERVER_INVALID_404_PATH, url),
        )

    if not (file_good or path_good):
        output.norm(
            "Site does not respond properly to non-existent file/path requests; skipping some checks."
        )

    if file_good:
        links, results = special_files.check_special_files(url)
        if len(results) > 0:
            reporter.display_results(results, "\t")

            new_files += links

        if args.files:
            output.empty()
            output.norm("Searching for common files (this will take a few minutes)...")

            with Spinner():
                try:
                    links, results = file_search.find_files(url)
                except Exception as error:
                    output.debug_exception()
                    output.error(f"Error running scan: {str(error)}")
                    results = None
                    links = None

            if results is not None and len(results) > 0:
                reporter.display_results(results, "\t")

            if links is not None and len(links) > 0:
                new_files += links

                for l in links:
                    if l not in orig_links:
                        output.norm(f"\tNew file found: {l}")

                output.empty()

    if path_good:
        links, results = special_files.check_special_paths(url)

        if len(results) > 0:
            reporter.display_results(results, "\t")

            new_files += links

        if args.dir:
            output.empty()
            output.norm(
                "Searching for common directories (this will take a few minutes)..."
            )

            with Spinner():
                try:
                    links, results = file_search.find_directories(
                        url, args.dirlistredir, args.dirrecursive
                    )
                except Exception as error:
                    output.debug_exception()
                    output.error(f"Error running scan: {str(error)}")
                    results = None
                    links = None

            if results is not None and len(results) > 0:
                reporter.display_results(results, "\t")

            if links is not None and len(links) > 0:
                new_files += links

                for l in links:
                    if l not in orig_links:
                        output.norm(f"\tNew directory found: {l}")

                output.empty()

    return new_files



import json
import os
import time
import zipfile
from typing import Dict, List, cast, Optional, Any, Union
from zipfile import ZipFile

from yawast.external.memory_size import Size
from yawast.reporting.enums import Vulnerabilities, Severity
from yawast.reporting.issue import Issue
from yawast.scanner.plugins.result import Result
from yawast.shared import output

_issues: Dict[str, Dict[Vulnerabilities, List[Issue]]] = {}
_info: Dict[str, Any] = {}
_domain: str = ""
_output_file: str = ""


def init(output_file: Union[str, None] = None) -> None:
    global _output_file

    if output_file is not None:
        # if we have something, let's figure out what
        output_file = os.path.abspath(output_file)
        if os.path.isdir(output_file):
            # it's a directory, so we are going to create a name
            name = f"yawast_{int(time.time())}.json"
            output_file = os.path.join(output_file, name)
        elif os.path.isfile(output_file) or os.path.isfile(f"{_output_file}.zip"):
            # the file already exists
            print("WARNING: Output file already exists; it will be replaced.")

        _output_file = output_file


def save_output():
    global _issues, _info, _output_file

    vulns = {}
    for vuln in Vulnerabilities:
        vulns[vuln.name] = {"severity": vuln.severity, "description": vuln.description}

    data = {
        "info": _convert_keys(_info),
        "issues": _convert_keys(_issues),
        "vulnerabilities": vulns,
    }
    json_data = json.dumps(data, sort_keys=True, indent=4)

    try:
        zf = ZipFile(f"{_output_file}.zip", "x", zipfile.ZIP_LZMA)
        zf.writestr(f"{os.path.basename(_output_file)}", json_data)
        zf.close()

        orig = "{0:cM}".format(Size(len(json_data)))
        comp = "{0:cM}".format(Size(os.path.getsize(f"{_output_file}.zip")))
        print(f"Saving {_output_file}.zip... (size reduced from {orig} to {comp})")
    except Exception as error:
        print(f"Error writing output file: {error}")


def get_output_file() -> str:
    global _output_file

    if len(_output_file) > 0:
        return f"{_output_file}.zip"
    else:
        return ""


def setup(domain: str) -> None:
    global _domain, _issues

    _domain = domain

    _issues[_domain] = {}


def is_registered(vuln: Vulnerabilities) -> bool:
    global _issues, _domain

    if _issues is None:
        return False
    else:
        if _domain in _issues:
            if _issues[_domain].get(vuln) is None:
                return False
            else:
                return True
        else:
            return False


def register_info(key: str, value: Any):
    global _info

    _info[key] = value


def register(issue: Issue) -> None:
    global _issues, _domain

    # make sure the Dict for _domain exists - this shouldn't normally be an issue, but is for unit tests
    if _domain not in _issues:
        _issues[_domain] = {}

    # if we haven't handled this issue yet, create a List for it
    if not is_registered(issue.vulnerability):
        _issues[_domain][issue.vulnerability] = []

    # we need to check to see if we already have this issue, for this URL, so we don't create dups
    # TODO: This isn't exactly efficient - refactor
    findings = _issues[_domain][issue.vulnerability]
    findings = cast(List[Issue], findings)
    for finding in findings:
        if finding.url == issue.url and finding.evidence == issue.evidence:
            # just bail out
            output.debug(f"Duplicate Issue: {issue.id} (duplicate of {finding.id})")

            return

    _issues[_domain][issue.vulnerability].append(issue)
    output.debug(
        f"Issue Registered: {issue.id} - {issue.vulnerability.name} - {issue.url}"
    )


def display(msg: str, issue: Issue) -> None:
    if issue.vulnerability.display_all or not is_registered(issue.vulnerability):
        if issue.severity == Severity.CRITICAL or issue.severity == Severity.HIGH:
            output.vuln(msg)
        elif issue.severity == Severity.MEDIUM or issue.severity == Severity.LOW:
            output.warn(msg)
        else:
            output.info(msg)

    # if there's no evidence, default to the msg - better than nothing
    if issue.evidence is None:
        issue.evidence = msg.strip()

    register(issue)


def display_results(results: List[Result], padding: Optional[str] = ""):
    for res in results:
        iss = Issue.from_result(res)
        display(f"{padding}{res.message}", iss)


def _convert_keys(dct: Dict) -> Dict:
    ret = {}

    for k, v in dct.items():
        if type(k) is Vulnerabilities:
            k = k.name

        if type(v) is dict:
            v = _convert_keys(v)

        ret[k] = v

    return ret



from typing import List

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import response_scanner
from yawast.scanner.plugins.result import Result
from yawast.shared import network


def check_cve_2019_5418(url: str) -> List[Result]:
    # this only applies to controllers, so skip the check unless the link ends with '/'
    if not url.endswith("/"):
        return []

    results = []

    res = network.http_get(
        url, False, {"Accept": "../../../../../../../../../etc/passwd{{"}
    )
    body = res.text
    req = network.http_build_raw_request(res.request)

    results += response_scanner.check_response(url, res)

    if "root:" in body:
        results.append(
            Result(
                f"Rails CVE-2019-5418: File Content Disclosure: {url}",
                Vulnerabilities.SERVER_RAILS_CVE_2019_5418,
                url,
                [body, req],
            )
        )

    return results



import uuid
from typing import Union, List

from yawast.reporting.enums import Vulnerabilities
from yawast.shared import output


class Result:
    def __init__(
        self,
        msg: str,
        vuln: Vulnerabilities,
        url: str,
        evidence: Union[str, List[str], None] = None,
    ):
        self.message = msg
        self.vulnerability = vuln
        self.url = url

        if evidence is not None:
            self.evidence = evidence

            # if the evidence is a string, lets tack on the message as an extra element
            if type(evidence) is str:
                self.evidence = [evidence, msg]
        else:
            # fall back to the message if we don't have evidence - better than nothing
            self.evidence = msg

        self.id = uuid.uuid4().hex

        output.debug(
            f"Result Created: {self.id} - {self.vulnerability.name} - {self.url}"
        )

    def __repr__(self):
        return f"Result: {self.id} - {self.vulnerability.name} - {self.url} - {self.message}"



from typing import List, Dict

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.result import Result


def get_waf(headers: Dict, raw: str, url: str) -> List[Result]:
    results = []

    if "Server" in headers:
        if headers["Server"] == "cloudflare":
            results.append(
                Result(
                    "WAF Detected: Cloudflare", Vulnerabilities.WAF_CLOUDFLARE, url, raw
                )
            )

    if "X-CDN" in headers or "X-Iinfo" in headers:
        results.append(
            Result("WAF Detected: Incapsula", Vulnerabilities.WAF_INCAPSULA, url, raw)
        )

    return results



from time import sleep
import sys
from typing import Optional

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from yawast.reporting import reporter, issue
from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.ssl import cert_info
from yawast.scanner.plugins.ssl_labs import api
from yawast.shared import output


def scan(url, args, domain):
    tty = sys.stdout.isatty()

    output.norm("Beginning SSL Labs scan (this could take a minute or two)")
    output.empty()

    # list the messages from SSL Labs - this is required as part of the ToS
    messages = api.get_info_message()
    for msg in messages:
        output.norm("[SSL Labs] {msg}".format(msg=msg))

    api.start_scan(domain)
    status = ""

    error_count = 0
    completed = []
    body = None

    while status != "READY" and status != "ERROR" and status != "DNS":
        sleep(5)

        try:
            status, body = api.check_scan(domain)
        except Exception:
            # if we find ourselves here, we want to try a couple more times before we give up for good
            output.debug_exception()

            if error_count > 3:
                raise
            else:
                error_count += 1

        if tty:
            # clear the current line
            sys.stdout.write("\r\033[K")

            # display the current status
            if "endpoints" in body:
                msg = ""

                for ep in body["endpoints"]:
                    if (
                        ep["statusMessage"] == "Ready"
                        and ep["ipAddress"] not in completed
                    ):
                        completed.append(ep["ipAddress"])

                        sys.stdout.write(
                            f'\r       Status - {ep["ipAddress"]}: {ep["statusMessage"]}\n\r'
                        )
                    elif (
                        ep["statusMessage"] != "Pending"
                        and ep["statusMessage"] != "Ready"
                    ):
                        # get the completion percentage
                        pct = "--"
                        if "progress" in ep:
                            pct = ep["progress"]

                        sm = "unknown"
                        if "statusDetailsMessage" in ep:
                            sm = ep["statusDetailsMessage"]

                        msg = f'\r       Status - {ep["ipAddress"]}: {ep["statusMessage"]} ({sm}) - {pct}%'

                        break

                if msg != "":
                    sys.stdout.write(msg)
                else:
                    sys.stdout.write(f"\r       Status: Working...")
            elif "status" in body:
                sys.stdout.write(f'\r       Status: {body["status"]}')
            else:
                sys.stdout.write(f"\r       Status: Working...")

            # flush the buffer, to make sure it's actually written
            sys.stdout.flush()
        else:
            print(".", end="", flush=True)

    output.empty()
    output.empty()

    # HACK: this needs to be refactored, once we have a better way to do it. This is awful.
    if not body.get("endpoints") is None:
        for ep in body["endpoints"]:
            output.norm(f'IP: {ep["ipAddress"]} - Grade: {ep["grade"]}')

            if ep["statusMessage"] == "Ready":
                _get_cert_info(body, ep, url)
                _get_protocol_info(ep, url)
                _get_vulnerability_info(ep, url)
            else:
                output.error(
                    f'Error getting information for IP: {ep["ipAddress"]}: {ep["statusMessage"]}'
                )
                output.empty()
    else:
        output.debug(f"Invalid response received: {body}")
        output.error("Invalid response received: Endpoint data not found.")
        output.empty()


def _get_cert_info(body, ep, url):
    # get the ChainCert info for the server cert - needed for extra details
    cert = None
    x509_cert = None

    for c in body["certs"]:
        if c["id"] == ep["details"]["certChains"][0]["certIds"][0]:
            cert = c
            x509_cert = x509.load_pem_x509_certificate(
                bytes(cert["raw"], "utf_8"), default_backend()
            )

    output.norm("Certificate Information:")

    if cert["issues"] > 0:
        output.warn("\tCertificate Has Issues - Not Valid")

        if cert["issues"] & 1 != 0:
            reporter.display(
                "\tCertificate Issue: no chain of trust",
                issue.Issue(Vulnerabilities.TLS_CERT_NO_TRUST, url),
            )

        if cert["issues"] & (1 << 1) != 0:
            reporter.display(
                "\tCertificate Issue: certificate not yet valid",
                issue.Issue(Vulnerabilities.TLS_CERT_NOT_YET_VALID, url),
            )

        if cert["issues"] & (1 << 2) != 0:
            reporter.display(
                "\tCertificate Issue: certificate expired",
                issue.Issue(Vulnerabilities.TLS_CERT_EXPIRED, url),
            )

        if cert["issues"] & (1 << 3) != 0:
            reporter.display(
                "\tCertificate Issue: hostname mismatch",
                issue.Issue(Vulnerabilities.TLS_CERT_HOSTNAME_MISMATCH, url),
            )

        if cert["issues"] & (1 << 4) != 0:
            reporter.display(
                "\tCertificate Issue: revoked",
                issue.Issue(Vulnerabilities.TLS_CERT_REVOKED, url),
            )

        if cert["issues"] & (1 << 5) != 0:
            reporter.display(
                "\tCertificate Issue: bad common name",
                issue.Issue(Vulnerabilities.TLS_CERT_BAD_COMMON_NAME, url),
            )

        if cert["issues"] & (1 << 6) != 0:
            reporter.display(
                "\tCertificate Issue: self-signed",
                issue.Issue(Vulnerabilities.TLS_CERT_SELF_SIGNED, url),
            )

        if cert["issues"] & (1 << 7) != 0:
            reporter.display(
                "\tCertificate Issue: blacklisted",
                issue.Issue(Vulnerabilities.TLS_CERT_BLACKLISTED, url),
            )

        if cert["issues"] & (1 << 8) != 0:
            reporter.display(
                "\tCertificate Issue: insecure signature",
                issue.Issue(Vulnerabilities.TLS_CERT_INSECURE_SIGNATURE, url),
            )

        if cert["issues"] & (1 << 9) != 0:
            reporter.display(
                "\tCertificate Issue: insecure key",
                issue.Issue(Vulnerabilities.TLS_CERT_INSECURE_KEY, url),
            )

        output.empty()

    output.norm(f'\tSubject: {cert["subject"]}')
    output.norm(f'\tCommon Names: {" ".join(cert["commonNames"])}')

    output.norm("\tAlternative names:")
    for name in cert["altNames"]:
        output.norm(f"\t\t{name}")

    output.norm(f'\tNot Before: {x509_cert.not_valid_before.isoformat(" ")}')
    output.norm(f'\tNot After: {x509_cert.not_valid_after.isoformat(" ")}')

    if cert["keyAlg"] == "EC":
        output.norm(
            f'\tKey: {cert["keyAlg"]} {cert["keySize"]} (RSA equivalent: {cert["keyStrength"]})'
        )
    else:
        if cert["keySize"] < 2048:
            output.vuln(f'\tKey: {cert["keyAlg"]} {cert["keySize"]}')
        else:
            output.norm(f'\tKey: {cert["keyAlg"]} {cert["keySize"]}')

    # TODO: Public Key Hash

    serial = format(x509_cert.serial_number, "02x")
    output.norm(f"\tSerial: {serial}")

    output.norm(f'\tIssuer: {cert["issuerSubject"]}')

    if "validationType" in cert:
        if cert["validationType"] == "E":
            output.norm("\tExtended Validation: Yes")
        elif cert["validationType"] == "D":
            output.norm("\tExtended Validation: No (Domain Control)")
        else:
            output.norm(f'\tExtended Validation: No ({cert["validationType"]})')

    if cert["sct"]:
        # check the first bit, SCT in cert
        if ep["details"]["hasSct"] & 1 != 0:
            output.norm("\tCertificate Transparency: SCT in certificate")

        # check second bit, SCT in stapled OSCP response
        if ep["details"]["hasSct"] & (1 << 1) != 0:
            output.norm("\tCertificate Transparency: SCT in the stapled OCSP response")

        # check third bit, SCT in the TLS extension
        if ep["details"]["hasSct"] & (1 << 2) != 0:
            output.norm(
                "\tCertificate Transparency: SCT in the TLS extension (ServerHello)"
            )
    else:
        output.norm("\tCertificate Transparency: No")

    output.norm(f'\tOCSP Must Staple: {cert["mustStaple"]}')

    if cert["revocationInfo"] & 1 != 0:
        output.norm("\tRevocation information: CRL information available")

    if cert["revocationInfo"] & (1 << 1) != 0:
        output.norm("\tRevocation information: OCSP information available")

    if cert["revocationStatus"] == 0:
        output.norm("\tRevocation Status: not checked")
    elif cert["revocationStatus"] == 1:
        output.vuln('\tRevocation Status: certificate revoked"')
    elif cert["revocationStatus"] == 2:
        output.norm("\tRevocation Status: certificate not revoked")
    elif cert["revocationStatus"] == 3:
        output.warn("\tRevocation Status: revocation check error")
    elif cert["revocationStatus"] == 4:
        output.warn("\tRevocation Status: no revocation information")
    elif cert["revocationStatus"] == 5:
        output.error("\tRevocation Status: SSL Labs internal error")
    else:
        output.error(
            f'\tRevocation Status: Unknown response ({cert["revocationStatus"]})'
        )

    if "crlRevocationStatus" in cert:
        if cert["crlRevocationStatus"] == 0:
            output.norm("\tCRL Revocation Status: not checked")
        elif cert["crlRevocationStatus"] == 1:
            output.vuln('\tCRL Revocation Status: certificate revoked"')
        elif cert["crlRevocationStatus"] == 2:
            output.norm("\tCRL Revocation Status: certificate not revoked")
        elif cert["crlRevocationStatus"] == 3:
            output.warn("\tCRL Revocation Status: revocation check error")
        elif cert["crlRevocationStatus"] == 4:
            output.warn("\tCRL Revocation Status: no revocation information")
        elif cert["crlRevocationStatus"] == 5:
            output.error("\tCRL Revocation Status: SSL Labs internal error")
        else:
            output.error(
                f'\tCRL Revocation Status: Unknown response ({cert["crlRevocationStatus"]})'
            )
    else:
        output.norm("\tCRL Revocation Status: Not Provided")

    if "ocspRevocationStatus" in cert:
        if cert["ocspRevocationStatus"] == 0:
            output.norm("\tOCSP Revocation Status: not checked")
        elif cert["ocspRevocationStatus"] == 1:
            output.vuln('\tOCSP Revocation Status: certificate revoked"')
        elif cert["ocspRevocationStatus"] == 2:
            output.norm("\tOCSP Revocation Status: certificate not revoked")
        elif cert["ocspRevocationStatus"] == 3:
            output.warn("\tOCSP Revocation Status: revocation check error")
        elif cert["ocspRevocationStatus"] == 4:
            output.warn("\tOCSP Revocation Status: no revocation information")
        elif cert["ocspRevocationStatus"] == 5:
            output.error("\tOCSP Revocation Status: SSL Labs internal error")
        else:
            output.error(
                f'\tOCSP Revocation Status: Unknown response ({cert["ocspRevocationStatus"]})'
            )
    else:
        output.norm("\tOCSP Revocation Status: Not Provided")

    output.empty()

    exts = cert_info.format_extensions(x509_cert)
    for ext in exts:
        output.norm(f"\tExtensions: {ext}")

    output.empty()

    scts = cert_info.get_scts(x509_cert)
    for sct in scts:
        output.norm(
            f'\tSCT: {cert_info.get_ct_log_name(sct[1])} - {sct[2].isoformat(" ")}'
        )

    output.empty()

    cert_hash = bytes.hex(x509_cert.fingerprint(hashes.SHA1()))
    output.norm(f"\tFingerprint: {cert_hash}")
    output.norm(f"\t\thttps://censys.io/certificates?q={cert_hash}")
    output.norm(f"\t\thttps://crt.sh/?q={cert_hash}")

    output.empty()

    output.norm("\tCertificate Chains:")
    for chain in ep["details"]["certChains"]:
        path_count = 0

        # build list of trust paths
        trust_paths = {}
        for path in chain["trustPaths"]:
            trusts = None

            # in practice, it seems there is only only per path, but just in case
            for trust in path["trust"]:
                trust_line = None

                if trust["isTrusted"]:
                    trust_line = f'{trust["rootStore"]} (trusted)'
                else:
                    trust_line = f'{trust["rootStore"]} ({trust["trustErrorMessage"]})'

                if trusts is None:
                    trusts = trust_line
                else:
                    trusts += f" {trust_line}"

            if trust_paths.get(tuple(path["certIds"])) is not None:
                trust_paths[tuple(path["certIds"])] += f" {trusts}"
            else:
                trust_paths[tuple(path["certIds"])] = trusts

        # process each of the trust paths
        for key in trust_paths.keys():
            path_count += 1

            output.norm(f"\t  Path {path_count}:")
            output.norm(f"\t   Root Stores: {trust_paths[key]}")

            if chain["issues"] & (1 << 1) != 0:
                output.warn("\tCertificate Chain Issue: incomplete chain")

            if chain["issues"] & (1 << 2) != 0:
                output.warn(
                    "\tCertificate Chain Issue: chain contains unrelated/duplicate certificates"
                )

            if chain["issues"] & (1 << 3) != 0:
                output.warn("\tCertificate Chain Issue: incorrect order")

            if chain["issues"] & (1 << 4) != 0:
                output.warn("\tCertificate Chain Issue: contains anchor")

            if cert["issues"] & (1 << 5) != 0:
                output.warn("\tCertificate Chain Issue: untrusted")

            for path_cert in key:
                for c in body["certs"]:
                    if c["id"] == path_cert:
                        output.norm(f'\t\t{c["subject"]}')
                        output.norm(
                            f'\t\t Signature: {c["sigAlg"]}  Key: {c["keyAlg"]}-{c["keySize"]}'
                        )

                        if cert_info.check_symantec_root(c["sha256Hash"]):
                            reporter.display(
                                "\t\t Untrusted Symantec Root",
                                issue.Issue(
                                    Vulnerabilities.TLS_SYMANTEC_ROOT,
                                    url,
                                    c["sha1Hash"],
                                ),
                            )

                        output.norm(f'\t\t  https://crt.sh/?q={c["sha1Hash"]}')

                        if c["sha256Hash"] not in chain["certIds"]:
                            output.norm("\t\t  (provided by server)")

    output.empty()


def _get_protocol_info(ep, url):
    output.norm("Configuration Information:")
    output.norm("\tProtocol Support:")

    # check protocols
    protos = {}
    tls13_enabled = False

    for proto in ep["details"]["protocols"]:
        if proto["name"] == "SSL":
            # show a vuln for SSLvX

            reporter.display(
                f'\t\t{proto["name"]} {proto["version"]}',
                issue.Issue(Vulnerabilities.TLS_LEGACY_SSL_ENABLED, url),
            )
        elif proto["name"] == "TLS" and proto["version"] == "1.0":
            # show a warn for TLSv1.0

            reporter.display(
                f'\t\t{proto["name"]} {proto["version"]}',
                issue.Issue(Vulnerabilities.TLS_VERSION_1_0_ENABLED, url),
            )
        elif proto["name"] == "TLS" and proto["version"] == "1.3":
            # capture TLS 1.3 status
            tls13_enabled = True

            output.norm(f'\t\t{proto["name"]} {proto["version"]}')
        else:
            output.norm(f'\t\t{proto["name"]} {proto["version"]}')

        protos[proto["id"]] = f'{proto["name"]} {proto["version"]}'

    if not tls13_enabled:
        reporter.display(
            "\t\tTLS 1.3 Is Not Enabled",
            issue.Issue(Vulnerabilities.TLS_VERSION_1_3_NOT_ENABLED, url),
        )

    output.empty()

    if "namedGroups" in ep["details"]:
        output.norm("\tNamed Group Support:")

        for group in ep["details"]["namedGroups"]["list"]:
            output.norm(f'\t\t{group["name"]} {group["bits"]}')

        output.empty()

    if "suites" in ep["details"]:
        output.norm("\tCipher Suite Support:")

        for proto_suites in ep["details"]["suites"]:
            output.norm(f'\t\t{protos[proto_suites["protocol"]]}')

            for suite in proto_suites["list"]:
                ke = _get_key_exchange(suite)

                strength = suite["cipherStrength"]

                if "3DES" in suite["name"]:
                    # in this case, the effective strength is only 112 bits,
                    #  which is what we want to report. So override SSL Labs
                    strength = 112

                if ke is not None:
                    suite_info = f'{suite["name"].ljust(50)} - {strength}-bits - {ke}'
                else:
                    suite_info = f'{suite["name"].ljust(50)} - {strength}-bits'

                if _is_cipher_suite_secure(suite):
                    if "CBC" in suite["name"]:
                        output.info(f"\t\t  {suite_info}")

                        reporter.register(
                            issue.Issue(
                                Vulnerabilities.TLS_CBC_CIPHER_SUITE, url, suite_info
                            )
                        )
                    else:
                        output.norm(f"\t\t  {suite_info}")
                else:
                    output.vuln(f"\t\t  {suite_info}")

                    reporter.register(
                        issue.Issue(
                            Vulnerabilities.TLS_INSECURE_CIPHER_SUITE, url, suite_info
                        )
                    )

    output.empty()

    _get_simulations(ep, protos)


def _get_key_exchange(suite, is_sim: Optional[bool] = False):
    ke = None

    if "kxType" in suite:
        if "namedGroupBits" in suite:
            ke = (
                f'{suite["kxType"]}-{suite["namedGroupBits"]} / '
                f'{suite["namedGroupName"]} ({suite["kxStrength"]} equivalent)'
            )
        else:
            if is_sim:
                if suite["kxType"] == "DH" and "dhBits" in suite:
                    ke = f'{suite["kxType"]}-{suite["dhBits"]}'
                else:
                    ke = f'{suite["kxType"]}-{suite["kxStrength"]}'
            else:
                ke = f'{suite["kxType"]}-{suite["kxStrength"]}'

    return ke


def _is_cipher_suite_secure(suite):
    secure = True

    # check for weak DH
    if "kxStrength" in suite and suite["kxStrength"] < 2048:
        secure = False

    # check for RC4
    if "RC4" in suite["name"]:
        secure = False

    # check for weak suites
    if suite["cipherStrength"] < 128:
        secure = False

    return secure


def _get_simulations(ep, protos):
    if ep["details"]["sims"]["results"] is not None:
        output.norm("\tHandshake Simulation:")

        for sim in ep["details"]["sims"]["results"]:
            name = f'{sim["client"]["name"]} {sim["client"]["version"]}'

            if "platform" in sim["client"]:
                name += f' / {sim["client"]["platform"]}'

            name = name.ljust(28)

            if sim["errorCode"] == 0:
                protocol = protos[sim["protocolId"]]
                ke = _get_key_exchange(sim, True)

                if ke is not None:
                    suite_name = f'{sim["suiteName"]} - {ke}'
                else:
                    suite_name = f'{sim["suiteName"]}'

                output.norm(f"\t\t{name} - {protocol} - {suite_name}")
            else:
                output.info(f"\t\t{name} - Simulation Failed")

        output.empty()


def _get_vulnerability_info(ep, url):
    output.norm("\tProtocol & Vulnerability Information:")

    if "sniRequired" in ep["details"]:
        if ep["details"]["sniRequired"]:
            output.info("\t\tSNI Required: Yes")
        else:
            output.norm("\t\tSNI Required: No")

    if "drownVulnerable" in ep["details"]:
        if ep["details"]["drownVulnerable"]:
            output.vuln("\t\tDROWN: Vulnerable")

            servers = ""
            for dh in ep["details"]["drownHosts"]:
                servers += f'({dh["ip"]}:{dh["port"]} - {dh["status"]}) '
                output.norm(f'\t\t\t{dh["ip"]}:{dh["port"]} - {dh["status"]}')
                output.norm(f'\t\t\t  https://test.drownattack.com/?site={dh["ip"]}')

            reporter.register(issue.Issue(Vulnerabilities.TLS_DROWN, url, servers))
        else:
            output.norm("\t\tDROWN: No")
    else:
        output.error("\t\tDROWN: Information Not Received")

    if "zeroRTTEnabled" in ep["details"]:
        if ep["details"]["zeroRTTEnabled"] == -2:
            output.error("\t\tTLS 1.3 0-RTT Support: Test Failed")
        elif ep["details"]["zeroRTTEnabled"] == -1:
            output.norm("\t\tTLS 1.3 0-RTT Support: Test Not Performed")
        elif ep["details"]["zeroRTTEnabled"] == 0:
            output.norm("\t\tTLS 1.3 0-RTT Support: No")
        elif ep["details"]["zeroRTTEnabled"] == 1:
            reporter.display(
                "\t\tTLS 1.3 0-RTT Support: Yes",
                issue.Issue(Vulnerabilities.TLS_VERSION_1_3_EARLY_DATA_ENABLED, url),
            )
        else:
            output.error(
                f'\t\tTLS 1.3 0-RTT Support: Unknown ({ep["details"]["zeroRTTEnabled"]})'
            )
    else:
        output.error("\t\tTLS 1.3 0-RTT Support: Information Not Received")

    if "renegSupport" in ep["details"]:
        if ep["details"]["renegSupport"] & 1 != 0:
            reporter.display(
                "\t\tSecure Renegotiation: insecure client-initiated renegotiation supported",
                issue.Issue(Vulnerabilities.TLS_INSECURE_RENEG, url),
            )

        if ep["details"]["renegSupport"] & (1 << 1) != 0:
            output.norm("\t\tSecure Renegotiation: secure renegotiation supported")

        if ep["details"]["renegSupport"] & (1 << 2) != 0:
            output.norm(
                "\t\tSecure Renegotiation: secure client-initiated renegotiation supported"
            )

        if ep["details"]["renegSupport"] & (1 << 3) != 0:
            output.norm(
                '\t\tSecure Renegotiation: server requires secure renegotiation support"'
            )
    else:
        output.error("\t\tSecure Renegotiation: Information Not Received")

    if "poodle" in ep["details"]:
        if ep["details"]["poodle"]:
            reporter.display(
                "\t\tPOODLE (SSL): Vulnerable",
                issue.Issue(Vulnerabilities.TLS_LEGACY_SSL_POODLE, url),
            )
        else:
            output.norm("\t\tPOODLE (SSL): No")
    else:
        output.error("\t\tPOODLE (SSL): Information Not Received")

    if "zombiePoodle" in ep["details"]:
        if ep["details"]["zombiePoodle"] == -1:
            output.error("\t\tZombie POODLE: Test Failed")
        elif ep["details"]["zombiePoodle"] == 0:
            output.error("\t\tZombie POODLE: Test Failed (Unknown)")
        elif ep["details"]["zombiePoodle"] == 1:
            output.norm("\t\tZombie POODLE: No")
        elif ep["details"]["zombiePoodle"] == 2:
            reporter.display(
                "\t\tZombie POODLE: Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_ZOMBIE_POODLE_NE, url),
            )
        elif ep["details"]["zombiePoodle"] == 3:
            output.vuln("\t\tZombie POODLE: Vulnerable - Exploitable")
            reporter.display(
                "\t\tZombie POODLE: Vulnerable - Exploitable",
                issue.Issue(Vulnerabilities.TLS_ZOMBIE_POODLE, url),
            )
        else:
            output.error(
                f'\t\tZombie POODLE: Unknown Response ({ep["details"]["zombiePoodle"]})'
            )
    else:
        output.error("\t\tZombie POODLE: Information Not Received")

    if "goldenDoodle" in ep["details"]:
        if ep["details"]["goldenDoodle"] == -1:
            output.error("\t\tGOLDENDOODLE: Test Failed")
        elif ep["details"]["goldenDoodle"] == 0:
            output.error("\t\tGOLDENDOODLE: Test Failed (Unknown)")
        elif ep["details"]["goldenDoodle"] == 1:
            output.norm("\t\tGOLDENDOODLE: No")
        elif ep["details"]["goldenDoodle"] == 4:
            reporter.display(
                "\t\tGOLDENDOODLE: Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_GOLDENDOODLE_NE, url),
            )
        elif ep["details"]["goldenDoodle"] == 5:
            reporter.display(
                "\t\tGOLDENDOODLE: Vulnerable - Exploitable",
                issue.Issue(Vulnerabilities.TLS_GOLDENDOODLE, url),
            )
        else:
            output.error(
                f't\tGOLDENDOODLE: Unknown Response ({ep["details"]["goldenDoodle"]})'
            )
    else:
        output.error("\t\tGOLDENDOODLE: Information Not Received")

    if "zeroLengthPaddingOracle" in ep["details"]:
        if ep["details"]["zeroLengthPaddingOracle"] == -1:
            output.error(
                "\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): Test Failed"
            )
        elif ep["details"]["zeroLengthPaddingOracle"] == 0:
            output.error(
                "\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): Test Failed (Unknown)"
            )
        elif ep["details"]["zeroLengthPaddingOracle"] == 1:
            output.norm("\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): No")
        elif ep["details"]["zeroLengthPaddingOracle"] == 6:
            reporter.display(
                "\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_OPENSSL_CVE_2019_1559_NE, url),
            )
        elif ep["details"]["zeroLengthPaddingOracle"] == 7:
            reporter.display(
                "\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): Vulnerable - Exploitable",
                issue.Issue(Vulnerabilities.TLS_OPENSSL_CVE_2019_1559, url),
            )
        else:
            output.error(
                f"\t\tOpenSSL 0-Length Padding Oracle (CVE-2019-1559): Unknown Response"
                f' ({ep["details"]["zeroLengthPaddingOracle"]})'
            )
    else:
        output.error(
            "OpenSSL 0-Length Padding Oracle (CVE-2019-1559): Information Not Received"
        )

    if "sleepingPoodle" in ep["details"]:
        if ep["details"]["sleepingPoodle"] == -1:
            output.error("\t\tSleeping POODLE: Test Failed")
        elif ep["details"]["sleepingPoodle"] == 0:
            output.error("\t\tSleeping POODLE: Test Failed (Unknown)")
        elif ep["details"]["sleepingPoodle"] == 1:
            output.norm("\t\tSleeping POODLE: No")
        elif ep["details"]["sleepingPoodle"] == 10:
            reporter.display(
                "\t\tSleeping POODLE: Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_SLEEPING_POODLE_NE, url),
            )
        elif ep["details"]["sleepingPoodle"] == 11:
            output.vuln("\t\tSleeping POODLE: Vulnerable - Exploitable")
            reporter.display(
                "\t\tSleeping POODLE: Vulnerable - Exploitable",
                issue.Issue(Vulnerabilities.TLS_SLEEPING_POODLE, url),
            )
        else:
            output.error(
                f'\t\tSleeping POODLE: Unknown Response ({ep["details"]["sleepingPoodle"]})'
            )
    else:
        output.error("\t\tSleeping POODLE: Information Not Received")

    if "poodleTls" in ep["details"]:
        if ep["details"]["poodleTls"] == -3:
            output.info("\t\tPOODLE (TLS): Inconclusive (Timeout)")
        elif ep["details"]["poodleTls"] == -2:
            output.info("\t\tPOODLE (TLS): TLS Not Supported")
        elif ep["details"]["poodleTls"] == -1:
            output.error("\t\tPOODLE (TLS): Test Failed")
        elif ep["details"]["poodleTls"] == 0:
            output.error("\t\tPOODLE (TLS): Test Failed (Unknown)")
        elif ep["details"]["poodleTls"] == 1:
            output.norm("\t\tPOODLE (TLS): No")
        elif ep["details"]["poodleTls"] == 2:
            reporter.display(
                "\t\tPOODLE (TLS): Vulnerable",
                issue.Issue(Vulnerabilities.TLS_POODLE, url),
            )
        else:
            output.error(
                f'\t\tPOODLE (TLS): Unknown Response ({ep["details"]["poodleTls"]})'
            )
    else:
        output.error("\t\tPOODLE (TLS): Information Not Received")

    if "fallbackScsv" in ep["details"]:
        if ep["details"]["fallbackScsv"]:
            output.norm("\t\tDowngrade Prevention: Yes")
        else:
            reporter.display(
                "\t\tDowngrade Prevention: No",
                issue.Issue(Vulnerabilities.TLS_FALLBACK_SCSV_MISSING, url),
            )
    else:
        output.error("t\tDowngrade Prevention: Information Not Received")

    if "compressionMethods" in ep["details"]:
        if ep["details"]["compressionMethods"] & 1 != 0:
            reporter.display(
                "\t\tCompression: DEFLATE",
                issue.Issue(Vulnerabilities.TLS_COMPRESSION_ENABLED, url),
            )
        else:
            output.norm("\t\tCompression: No")
    else:
        output.error("\t\tCompression: Information Not Received")

    if "heartbeat" in ep["details"]:
        if ep["details"]["heartbeat"]:
            reporter.display(
                "\t\tHeartbeat: Enabled",
                issue.Issue(Vulnerabilities.TLS_HEARTBEAT_ENABLED, url),
            )
        else:
            output.norm("\t\tHeartbeat: Disabled")
    else:
        output.error("\t\tHeartbeat: Information Not Received")

    if "heartbleed" in ep["details"]:
        if ep["details"]["heartbleed"]:
            reporter.display(
                "\t\tHeartbleed: Vulnerable",
                issue.Issue(Vulnerabilities.TLS_HEARTBLEEDL, url),
            )
        else:
            output.norm("\t\tHeartbleed: No")
    else:
        output.error("\t\tHeartbleed: Information Not Received")

    if "ticketbleed" in ep["details"]:
        if ep["details"]["ticketbleed"] == -1:
            output.error("\t\tTicketbleed (CVE-2016-9244): Test Failed")
        elif ep["details"]["ticketbleed"] == 0:
            output.error("\t\tTicketbleed (CVE-2016-9244): Test Failed (Unknown)")
        elif ep["details"]["ticketbleed"] == 1:
            output.norm("\t\tTicketbleed (CVE-2016-9244): No")
        elif ep["details"]["ticketbleed"] == 2:
            reporter.display(
                "\t\tTicketbleed (CVE-2016-9244): Vulnerable",
                issue.Issue(Vulnerabilities.TLS_TICKETBLEED, url),
            )
        else:
            output.error(
                f'\t\tTicketbleed (CVE-2016-9244): Unknown Response ({ep["details"]["ticketbleed"]})'
            )
    else:
        output.error("\t\tTicketbleed (CVE-2016-9244): Information Not Received")

    if "openSslCcs" in ep["details"]:
        if ep["details"]["openSslCcs"] == -1:
            output.error("\t\tOpenSSL CCS (CVE-2014-0224): Test Failed")
        elif ep["details"]["openSslCcs"] == 0:
            output.error("\t\tOpenSSL CCS (CVE-2014-0224): Test Failed (Unknown)")
        elif ep["details"]["openSslCcs"] == 1:
            output.norm("\t\tOpenSSL CCS (CVE-2014-0224): No")
        elif ep["details"]["openSslCcs"] == 2:
            reporter.display(
                "\t\tOpenSSL CCS (CVE-2014-0224): Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_OPENSSL_CVE_2014_0224_NE, url),
            )
        elif ep["details"]["openSslCcs"] == 3:
            output.vuln("\t\tOpenSSL CCS (CVE-2014-0224): Vulnerable")
            reporter.display(
                "\t\tOpenSSL CCS (CVE-2014-0224): Vulnerable",
                issue.Issue(Vulnerabilities.TLS_OPENSSL_CVE_2014_0224, url),
            )
        else:
            output.error(
                f'\t\tOpenSSL CCS (CVE-2014-0224): Unknown Response ({ep["details"]["openSslCcs"]})'
            )
    else:
        output.error("\t\tOpenSSL CCS (CVE-2014-0224): Information Not Received")

    if "openSSLLuckyMinus20" in ep["details"]:
        if ep["details"]["openSSLLuckyMinus20"] == -1:
            output.error("\t\tOpenSSL Padding Oracle (CVE-2016-2107): Test Failed")
        elif ep["details"]["openSSLLuckyMinus20"] == 0:
            output.error(
                "\t\tOpenSSL Padding Oracle (CVE-2016-2107): Test Failed (Unknown)"
            )
        elif ep["details"]["openSSLLuckyMinus20"] == 1:
            output.norm("\t\tOpenSSL Padding Oracle (CVE-2016-2107): No")
        elif ep["details"]["openSSLLuckyMinus20"] == 2:
            reporter.display(
                "\t\tOpenSSL Padding Oracle (CVE-2016-2107): Vulnerable",
                issue.Issue(Vulnerabilities.TLS_OPENSSL_CVE_2016_2107, url),
            )
        else:
            output.error(
                f"\t\tOpenSSL Padding Oracle (CVE-2016-2107): Unknown Response "
                f'({ep["details"]["openSSLLuckyMinus20"]})'
            )
    else:
        output.error(
            "\t\tOpenSSL Padding Oracle (CVE-2016-2107): Information Not Received"
        )

    if "bleichenbacher" in ep["details"]:
        if ep["details"]["bleichenbacher"] == -1:
            output.error("\t\tROBOT: Test Failed")
        elif ep["details"]["bleichenbacher"] == 0:
            output.error("\t\tROBOT: Test Failed (Unknown)")
        elif ep["details"]["bleichenbacher"] == 1:
            output.norm("\t\tROBOT: No")
        elif ep["details"]["bleichenbacher"] == 2:
            reporter.display(
                "\t\tROBOT: Vulnerable - Not Exploitable",
                issue.Issue(Vulnerabilities.TLS_ROBOT_ORACLE_WEAK, url),
            )
        elif ep["details"]["bleichenbacher"] == 3:
            reporter.display(
                "\t\tROBOT: Vulnerable - Exploitable",
                issue.Issue(Vulnerabilities.TLS_ROBOT_ORACLE_STRONG, url),
            )
        elif ep["details"]["bleichenbacher"] == 4:
            output.norm("\t\tROBOT: Unknown - Inconsistent Results")
        else:
            output.error(
                f'\t\tROBOT: Unknown Response ({ep["details"]["bleichenbacher"]})'
            )
    else:
        output.error("\t\tROBOT: Information Not Received")

    if "forwardSecrecy" in ep["details"]:
        if ep["details"]["forwardSecrecy"] & (1 << 2) != 0:
            output.norm("\t\tForward Secrecy: Yes (all simulated clients)")
        elif ep["details"]["forwardSecrecy"] & (1 << 1) != 0:
            output.info("\t\tForward Secrecy: Yes (modern clients)")
        elif ep["details"]["forwardSecrecy"] & 1 != 0:
            reporter.display(
                "\t\tForward Secrecy: Yes (limited support)",
                issue.Issue(Vulnerabilities.TLS_LIMITED_FORWARD_SECRECY, url),
            )
        else:
            output.vuln("\t\tForward Secrecy: No")
    else:
        output.error("\t\tForward Secrecy: Information Not Received")

    if "supportsAead" in ep["details"]:
        if ep["details"]["supportsAead"]:
            output.norm("\t\tAEAD Cipher Suites Supported: Yes")
        else:
            reporter.display(
                "\t\tAEAD Cipher Suites Supported: No",
                issue.Issue(Vulnerabilities.TLS_NO_AEAD_SUPPORT, url),
            )
    else:
        output.error("\t\tAEAD Cipher Suites Supported: Information Not Received")

    if "supportsCBC" in ep["details"]:
        if ep["details"]["supportsCBC"]:
            output.info("\t\tCBC Cipher Suites Supported: Yes")
        else:
            output.norm("\t\tCBC Cipher Suites Supported: No")
    else:
        output.error("\t\tCBC Cipher Suites Supported: Information Not Received")

    if "alpnProtocols" in ep["details"]:
        output.norm(f'\t\tALPN: {ep["details"]["alpnProtocols"]}')

    if "npnProtocols" in ep["details"]:
        output.norm(f'\t\tNPN: {ep["details"]["npnProtocols"]}')

    if "sessionResumption" in ep["details"]:
        if ep["details"]["sessionResumption"] == 0:
            output.norm("\t\tSession Resumption: Not Enabled / Empty Tickets")
        elif ep["details"]["sessionResumption"] == 1:
            output.norm("\t\tSession Resumption: Enabled / No Resumption")
        elif ep["details"]["sessionResumption"] == 2:
            reporter.display(
                "\t\tSession Resumption: Enabled",
                issue.Issue(Vulnerabilities.TLS_SESSION_RESP_ENABLED, url),
            )
        else:
            output.error(
                f'\t\tSession Resumption: Unknown Response ({ep["details"]["sessionResumption"]})'
            )
    else:
        output.error("\t\tSession Resumption: Information Not Received")

    if "ocspStapling" in ep["details"]:
        if ep["details"]["ocspStapling"]:
            output.norm("\t\tOCSP Stapling: Yes")
        else:
            reporter.display(
                "\t\tOCSP Stapling: No",
                issue.Issue(Vulnerabilities.TLS_OCSP_STAPLE_MISSING, url),
            )
    else:
        output.error("\t\tOCSP Stapling: Information Not Received")

    if "miscIntolerance" in ep["details"]:
        if ep["details"]["miscIntolerance"] & 1 != 0:
            output.info("\t\tTLS Extension Intolerance: Yes")

        if ep["details"]["miscIntolerance"] & (1 << 1) != 0:
            output.warn("\t\tLong Handshake Intolerance: Yes")

        if ep["details"]["miscIntolerance"] & (1 << 2) != 0:
            output.warn("\t\tLong Handshake Intolerance: Workaround Success")

    if "protocolIntolerance" in ep["details"]:
        if ep["details"]["protocolIntolerance"] & 1 != 0:
            output.warn("\t\tProtocol Intolerance: TLS 1.0")

        if ep["details"]["protocolIntolerance"] & (1 << 1) != 0:
            output.warn("\t\tProtocol Intolerance: TLS 1.1")

        if ep["details"]["protocolIntolerance"] & (1 << 2) != 0:
            output.warn("\t\tProtocol Intolerance: TLS 1.2")

        if ep["details"]["protocolIntolerance"] & (1 << 3) != 0:
            output.warn("\t\tProtocol Intolerance: TLS 1.3")

        if ep["details"]["protocolIntolerance"] & (1 << 4) != 0:
            output.warn("\t\tProtocol Intolerance: TLS 1.152")

        if ep["details"]["protocolIntolerance"] & (1 << 5) != 0:
            output.warn("\t\tProtocol Intolerance: TLS 2.152")

    if "freak" in ep["details"]:
        if ep["details"]["freak"]:
            reporter.display(
                "\t\tFREAK: Vulnerable (512-bit key exchange supported)",
                issue.Issue(Vulnerabilities.TLS_FREAK, url),
            )
        else:
            output.norm("\t\tFREAK: No")
    else:
        output.error("\t\tFREAK: Information Not Received")

    if "logjam" in ep["details"]:
        if ep["details"]["logjam"]:
            reporter.display(
                "\t\tLogjam: Vulnerable", issue.Issue(Vulnerabilities.TLS_LOGJAM, url)
            )
        else:
            output.norm("\t\tLogjam: No")
    else:
        output.error("\t\tLogjam: Information Not Received")

    if "dhUsesKnownPrimes" in ep["details"]:
        if ep["details"]["dhUsesKnownPrimes"] == 0:
            output.norm("\t\tUses common DH primes: No")
        elif ep["details"]["dhUsesKnownPrimes"] == 1:
            output.warn("\t\tUses common DH primes: Yes (not weak)")
            reporter.display(
                "\t\tUses common DH primes: Yes (weak)",
                issue.Issue(Vulnerabilities.TLS_DH_KNOWN_PRIMES_STRONG, url),
            )
        elif ep["details"]["dhUsesKnownPrimes"] == 2:
            reporter.display(
                "\t\tUses common DH primes: Yes (weak)",
                issue.Issue(Vulnerabilities.TLS_DH_KNOWN_PRIMES_WEAK, url),
            )
        else:
            output.error(
                f'\t\tUses common DH primes: Unknown Response ({ep["details"]["dhUsesKnownPrimes"]})'
            )

    if "dhYsReuse" in ep["details"]:
        if ep["details"]["dhYsReuse"]:
            reporter.display(
                "\t\tDH public server param (Ys) reuse: Yes",
                issue.Issue(Vulnerabilities.TLS_DH_PARAM_REUSE, url),
            )
        else:
            output.norm("\t\tDH public server param (Ys) reuse: No")

    if "ecdhParameterReuse" in ep["details"]:
        if ep["details"]["ecdhParameterReuse"]:
            reporter.display(
                "\t\tECDH Public Server Param Reuse: Yes",
                issue.Issue(Vulnerabilities.TLS_ECDH_PARAM_REUSE, url),
            )
        else:
            output.norm("\t\tECDH Public Server Param Reuse: No")

    output.empty()



import inspect
from typing import List, Union

from bs4 import BeautifulSoup
from requests.models import Response

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import http_basic, retirejs, error_checker, http_utils
from yawast.scanner.plugins.http.servers import rails, apache_tomcat
from yawast.scanner.plugins.result import Result
from yawast.shared import network


def check_response(
    url: str, res: Response, soup: Union[BeautifulSoup, None] = None
) -> List[Result]:
    # make sure we actually have something
    if res is None:
        return []

    results = []

    raw_full = network.http_build_raw_response(res)

    if http_utils.is_text(res):
        body = res.text

        if soup is None:
            soup = BeautifulSoup(body, "html.parser")

        # check for things thar require parsed HTML
        results += retirejs.get_results(soup, url, raw_full)
        results += apache_tomcat.get_version(url, res)
        results += error_checker.check_response(url, res, body)

    results += http_basic.get_header_issues(res.headers, raw_full, url)
    results += http_basic.get_cookie_issues(res, raw_full, url)

    # this function will trigger a recursive call, as it calls this to check the response.
    # to deal with this, we'll check the caller, to make sure it's not what we're about to call.
    if "check_cve_2019_5418" not in inspect.stack()[1].function:
        results += rails.check_cve_2019_5418(url)

    results += _check_charset(url, res, raw_full)

    return results


def _check_charset(url: str, res: Response, raw: str) -> List[Result]:
    results = []

    # if the body is empty, we really don't care about this
    if len(res.content) == 0:
        return results

    if "Content-Type" in res.headers:
        content_type = str(res.headers["Content-Type"]).lower()

        if "charset" not in content_type and "text/html" in content_type:
            # not charset specified
            results.append(
                Result(
                    f"Charset Not Defined in '{res.headers['Content-Type']}' at {url}",
                    Vulnerabilities.HTTP_HEADER_CONTENT_TYPE_NO_CHARSET,
                    url,
                    [res.headers["Content-Type"], raw],
                )
            )
    else:
        # content-type missing
        results.append(
            Result(
                f"Content-Type Missing: {url} ({res.request.method} - {res.status_code})",
                Vulnerabilities.HTTP_HEADER_CONTENT_TYPE_MISSING,
                url,
                raw,
            )
        )

    return results



import socket
from urllib.parse import urlparse

from yawast.scanner.cli import dns
from yawast.shared import utils


def start(args, url):
    print(f"Scanning: {url}")

    # parse the URL, we'll need this
    parsed = urlparse(url)
    # get rid of any port number & credentials that may exist
    domain = utils.get_domain(parsed.netloc)

    # make sure it resolves
    try:
        socket.gethostbyname(domain)
    except socket.gaierror as error:
        print(f"Fatal Error: Unable to resolve {domain} ({str(error)})")

        return

    dns.scan(args, url, domain)



from setuptools import setup, find_packages
from os import path


root_path = path.abspath(path.dirname(__file__))


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass("yawast")


def get_long_description():
    """Convert the README file into the long description.
    """
    with open(path.join(root_path, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
    return long_description


setup(
    name="yawast",
    version=version,
    cmdclass=cmdclass,
    description="The YAWAST Antecedent Web Application Security Toolkit",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/adamcaudill/yawast",
    project_urls={
        "Bug Reports": "https://github.com/adamcaudill/yawast/issues",
        "Source": "https://github.com/adamcaudill/yawast",
        "Changelog": "https://github.com/adamcaudill/yawast/blob/master/CHANGELOG.md",
    },
    author="Adam Caudill",
    author_email="adam@adamcaudill.com",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    scripts=["bin/yawast"],
    install_requires=[
        "validator-collection",
        "requests",
        "publicsuffixlist",
        "dnspython",
        "urllib3",
        "colorama",
        "sslyze==2.1.3",
        "nassl",
        "cryptography==2.5",
        "packaging",
        "beautifulsoup4",
        "psutil",
        "pipenv",
    ],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    keywords="security tls ssl dns http scan vulnerability",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Natural Language :: English",
        "Topic :: Internet :: Name Service (DNS)",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Networking",
        "Topic :: Software Development :: Testing",
        "Topic :: Security",
    ],
)



import socket

from yawast.shared import network, utils, output
from yawast.scanner.cli import dns, ssl_labs, ssl_internal, ssl_sweet32, http
from urllib.parse import urlparse, urlunparse


def start(args, url):
    print(f"Scanning: {url}")

    # parse the URL, we'll need this
    parsed = urlparse(url)
    # get rid of any port number & credentials that may exist
    domain = utils.get_domain(parsed.netloc)

    # make sure it resolves
    try:
        socket.gethostbyname(domain)
    except socket.gaierror as error:
        print(f"Fatal Error: Unable to resolve {domain} ({str(error)})")

        return

    # perform some connection testing
    if parsed.scheme == "http":
        try:
            # check for TLS redirect
            tls_redirect = network.check_ssl_redirect(url)
            if tls_redirect != url:
                print(f"Server redirects to TLS: Scanning: {tls_redirect}")

                url = tls_redirect
                parsed = urlparse(url)
        except Exception:
            output.debug_exception()

            # we tried to connect to port 80, and it failed
            # this could mean a couple things, first, we need to
            #  see if it answers to 443
            parsed = parsed._replace(scheme="https")
            url = urlunparse(parsed)

            print("Server does not respond to HTTP, switching to HTTPS")
            print()
            print(f"Scanning: {url}")

            # grab the head, to see if we get anything
            try:
                network.http_head(url, timeout=5)

                print()
            except Exception as err:
                output.debug_exception()

                print(f"Fatal Error: Can not connect to {url} ({str(err)})")
                return
    else:
        # if we are scanning HTTPS, try HTTP to see what it does
        try:
            http_parsed = parsed._replace(scheme="http")
            http_url = urlunparse(http_parsed)

            network.http_head(http_url, timeout=5)

            print("Server responds to HTTP requests")
            print()
        except Exception:
            output.debug_exception()

            print("Server does not respond to HTTP requests")
            print()

    # check for www redirect
    www_redirect = network.check_www_redirect(url)
    if www_redirect is not None and www_redirect != url:
        print(f"Server performs WWW redirect: Scanning: {www_redirect}")
        url = www_redirect

    if not args.nodns:
        dns.scan(args, url, domain)

    # check to see if we are looking at an HTTPS server
    if parsed.scheme == "https" and not args.nossl:
        if args.internalssl or utils.is_ip(domain) or utils.get_port(url) != 443:
            # use internal scanner
            ssl_internal.scan(args, url, domain)
        else:
            try:
                ssl_labs.scan(args, url, domain)
            except Exception as error:
                output.debug_exception()

                output.error(f"Error running scan with SSL Labs: {str(error)}")

        if args.tdessessioncount:
            ssl_sweet32.scan(args, url, domain)

    http.scan(args, url, domain)

    # reset any stored data
    http.reset()

    return



import time
from multiprocessing import Manager, Lock
from multiprocessing.dummy import Pool
from typing import List, Tuple
from urllib.parse import urljoin

from bs4 import BeautifulSoup

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import response_scanner, http_utils
from yawast.scanner.plugins.result import Result
from yawast.shared import network, output

_links = []
_insecure = []
_lock = Lock()
_tasks = []


def spider(url) -> Tuple[List[str], List[Result]]:
    global _links, _insecure, _tasks, _lock

    results = []

    # create processing pool
    pool = Pool()
    mgr = Manager()
    queue = mgr.Queue()

    asy = pool.apply_async(_get_links, (url, [url], queue, pool))

    with _lock:
        _tasks.append(asy)

    while True:
        if all(t is None or t.ready() for t in _tasks):
            break
        else:
            count_none = 0
            count_ready = 0
            count_not_ready = 0

            for t in _tasks:
                if t is None:
                    count_none += 1
                elif t.ready():
                    count_ready += 1
                else:
                    count_not_ready += 1

            output.debug(
                f"Spider Task Status: None: {count_none}, Ready: {count_ready}, Not Ready: {count_not_ready}"
            )

        time.sleep(3)

    pool.close()

    while not queue.empty():
        res = queue.get()

        if len(res) > 0:
            for re in res:
                if re not in results:
                    results.append(re)

    # copy data and reset
    links = _links[:]
    _links = []
    _insecure = []

    return links, results


def _get_links(base_url: str, urls: List[str], queue, pool):
    global _links, _insecure, _tasks, _lock

    results = []

    for url in urls:
        try:
            to_process = []

            res = network.http_get(url, False)

            if http_utils.is_text(res):
                soup = BeautifulSoup(res.text, "html.parser")
            else:
                # no clue what this is
                results += response_scanner.check_response(url, res)

                return

            # check the response for issues
            results += response_scanner.check_response(url, res, soup)

            for link in soup.find_all("a"):
                href = link.get("href")

                if href is not None:
                    # check to see if this link is in scope
                    if base_url in href and href not in _links:
                        if "." in href.split("/")[-1]:
                            file_ext = href.split("/")[-1].split(".")[-1]
                        else:
                            file_ext = None

                        with _lock:
                            _links.append(href)

                        # filter out some of the obvious binary files
                        if file_ext is None or file_ext not in [
                            "gzip",
                            "jpg",
                            "jpeg",
                            "gif",
                            "woff",
                            "zip",
                            "exe",
                            "gz",
                            "pdf",
                        ]:
                            to_process.append(href)
                        else:
                            output.debug(
                                f'Skipping URL "{href}" due to file extension "{file_ext}"'
                            )
                    else:
                        if (
                            "https://" in base_url
                            and "http://" in href
                            and href not in _insecure
                        ):
                            # link from secure to insecure
                            with _lock:
                                _insecure.append(href)

                            results.append(
                                Result(
                                    f"Insecure Link: {url} links to {href}",
                                    Vulnerabilities.HTTP_INSECURE_LINK,
                                    url,
                                    [href, res.text],
                                )
                            )
                        pass

            # handle redirects
            if "Location" in res.headers:
                redirect = res.headers["Location"]

                # check for relative link
                if str(redirect).startswith("/"):
                    redirect = urljoin(base_url, redirect)

                # make sure that we aren't redirected out of scope
                if base_url in redirect:
                    to_process.append(redirect)

            asy = pool.apply_async(_get_links, (base_url, to_process, queue, pool))

            with _lock:
                _tasks.append(asy)
        except Exception:
            output.debug_exception()

    queue.put(results)



import base64
import typing

from typing import List
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.x509.oid import NameOID

from yawast.shared import network

_ct_log_info = {}


def format_extensions(cert: x509.Certificate):
    extensions = []

    for ext in cert.extensions:
        name = ext.oid._name
        critical = ext.critical

        # exclude SUBJECT_ALTERNATIVE_NAME & PRECERT_SIGNED_CERTIFICATE_TIMESTAMPS
        if (
            ext.oid.dotted_string != "2.5.29.17"
            and ext.oid.dotted_string != "1.3.6.1.4.1.11129.2.4.2"
        ):
            data = ""
            for prop, value in vars(ext.value).items():
                if type(value) is bytes:
                    # convert byte strings to hex

                    value = value.hex()
                elif type(value) is list:
                    # some values are lists of more values, these need special processing.

                    item_data = []
                    for item in value:
                        if type(item) is x509.ObjectIdentifier:
                            # if this is an OID, we only care about the name

                            if item._name != "Unknown OID":
                                item_data.append(item._name)
                            else:
                                item_data.append(item.dotted_string)
                        elif type(item) is x509.PolicyInformation:
                            item_data.append(
                                "policy_identifier={oid}, policy_qualifiers={pq}".format(
                                    oid=item.policy_identifier.dotted_string,
                                    pq=item.policy_qualifiers,
                                )
                            )
                        elif type(item) is x509.DistributionPoint:
                            points = []

                            for point in item.full_name:
                                points.append(point.value)

                            item_data.append(" ".join(points).strip())
                        elif type(item) is x509.AccessDescription:
                            if item.access_method._name != "Unknown OID":
                                name = item.access_method._name
                            else:
                                name = item.access_method.dotted_string

                            item_data.append(
                                "{name}={url}".format(
                                    name=name, url=item.access_location.value
                                )
                            )
                        else:
                            item_data.append(str(item))

                    # assemble and clean up the data
                    value = "(" + ", ".join(item_data).rstrip().rstrip(",") + ")"

                data += "{prop}={val}, ".format(prop=prop.lstrip("_"), val=value)

            data = data.rstrip().rstrip(",")

            extensions.append(
                "{name}, critical={critical}, {value}".format(
                    name=name, critical=critical, value=data
                )
            )

    return extensions


def get_scts(cert: x509.Certificate):
    scts = []

    for ext in cert.extensions:
        # check for the PRECERT_SIGNED_CERTIFICATE_TIMESTAMPS OID
        if ext.oid.dotted_string == "1.3.6.1.4.1.11129.2.4.2":
            for sct in ext.value:
                scts.append(
                    [sct.version, bytes.hex(sct.log_id), sct.timestamp, sct.entry_type]
                )

    return scts


def get_common_names(cert: x509.Certificate) -> List[str]:
    names = []

    for name in cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME):
        names.append(name.value)

    return names


def get_alt_names(cert: x509.Certificate) -> List[str]:
    for ext in cert.extensions:
        # SUBJECT_ALTERNATIVE_NAME
        if ext.oid.dotted_string == "2.5.29.17":
            ext = typing.cast(x509.SubjectAlternativeName, ext)

            return ext.value.get_values_for_type(x509.DNSName)


def get_must_staple(cert: x509.Certificate):
    ret = False

    for ext in cert.extensions:
        # Must Staple OID
        if ext.oid.dotted_string == "1.3.6.1.5.5.7.1.24":
            ret = True

    return ret


def check_symantec_root(root_hash):
    # list of Symantec root cert hashes
    roots = [
        "08297a4047dba23680c731db6e317653ca7848e1bebd3a0b0179a707f92cf178",
        "2399561127a57125de8cefea610ddf2fa078b5c8067f4e828290bfb860e84b3c",
        "2834991cf677466d22baac3b0055e5b911d9a9e55f5b85ba02dc566782c30e8a",
        "2930bd09a07126bdc17288d4f2ad84645ec948607907a97b5ed0b0b05879ef69",
        "2f274e48aba4ac7b765933101775506dc30ee38ef6acd5c04932cfe041234220",
        "309b4a87f6ca56c93169aaa99c6d988854d7892bd5437e2d07b29cbeda55d35d",
        "3266967e59cd68008d9dd320811185c704205e8d95fdd84f1c7b311e6704fc32",
        "341de98b1392abf7f4ab90a960cf25d4bd6ec65b9a51ce6ed067d00ec7ce9b7f",
        "363f3c849eab03b0a2a0f636d7b86d04d3ac7fcfe26a0a9121ab9795f6e176df",
        "37d51006c512eaab626421f1ec8c92013fc5f82ae98ee533eb4619b8deb4d06c",
        "3a43e220fe7f3ea9653d1e21742eac2b75c20fd8980305bc502caf8c2d9b41a1",
        "3f9f27d583204b9e09c8a3d2066c4b57d3a2479c3693650880505698105dbce9",
        "44640a0a0e4d000fbd574d2b8a07bdb4d1dfed3b45baaba76f785778c7011961",
        "4b03f45807ad70f21bfc2cae71c9fde4604c064cf5ffb686bae5dbaad7fdd34c",
        "53dfdfa4e297fcfe07594e8c62d5b8ab06b32c7549f38a163094fd6429d5da43",
        "5b38bd129e83d5a0cad23921089490d50d4aae370428f8ddfffffa4c1564e184",
        "5edb7ac43b82a06a8761e8d7be4979ebf2611f7dd79bf91c1c6b566a219ed766",
        "5f0b62eab5e353ea6521651658fbb65359f443280a4afbd104d77d10f9f04c07",
        "614fd18da1490560cdad1196e2492ab7062eab1a67b3a30f1d0585a7d6ba6824",
        "69ddd7ea90bb57c93e135dc85ea6fcd5480b603239bdc454fc758b2a26cf7f79",
        "76ef4762e573206006cbc338b17ca4bc200574a11928d90c3ef31c5e803e6c6f",
        "83ce3c1229688a593d485f81973c0f9195431eda37cc5e36430e79c7a888638b",
        "87c678bfb8b25f38f7e97b336956bbcf144bbacaa53647e61a2325bc1055316b",
        "8d722f81a9c113c0791df136a2966db26c950a971db46b4199f4ea54b78bfb9f",
        "8dbb5a7c06c20ef62dd912a36740992ff6e1e8583d42ede257c3affd7c769399",
        "8f9e2751dcd574e9ba90e744ea92581fd0af640ae86ac1ce2198c90f96b44823",
        "92a9d9833fe1944db366e8bfae7a95b6480c2d6c6c2a1be65d4236b608fca1bb",
        "944554239d91ed9efedcf906d5e8113160b46fc816dc6bdc77b89da29b6562b9",
        "9acfab7e43c8d880d06b262a94deeee4b4659989c3d0caf19baf6405e41ab7df",
        "9d190b2e314566685be8a889e27aa8c7d7ae1d8aaddba3c1ecf9d24863cd34b9",
        "9e503738722e0a104cf659ff9f92f0b5b3662acd112d4664d1e7db93abf46a59",
        "a0234f3bc8527ca5628eec81ad5d69895da5680dc91d1cb8477f33f878b95b0b",
        "a0459b9f63b22559f5fa5d4c6db3f9f72ff19342033578f073bf1d1b46cbb912",
        "a4310d50af18a6447190372a86afaf8b951ffb431d837f1e5688b45971ed1557",
        "a4b6b3996fc2f306b3fd8681bd63413d8c5009cc4fa329c2ccf0e2fa1b140305",
        "b32396746453442f353e616292bb20bbaa5d23b546450fdb9c54b8386167d529",
        "b478b812250df878635c2aa7ec7d155eaa625ee82916e2cd294361886cd1fbd4",
        "bb6ce72f0e64bfd93ade14b1becf8c41e7bc927cafb477a3a95878c01aa26c3e",
        "bcff2ab03578ebbfb219b65e854cf26a3d8dfe6d1acf3e765b8636827b81eaee",
        "c38dcb38959393358691ea4d4f3ce495ce748996e64ed1891d897a0fc4dd55c6",
        "c4fa68f8270924c300cbc0d3615a7b88e82231749cf6522452272222c9f0a83e",
        "ca2d82a08677072f8ab6764ff035676cfe3e5e325e012172df3f92096db79b85",
        "cb627d18b58ad56dde331a30456bc65c601a4e9b18dedcea08e7daaa07815ff0",
        "cb6b05d9e8e57cd882b10b4db70de4bb1de42ba48a7bd0318b635bf6e7781a9d",
        "cbb02707160f4f77291a27561448691ca5901808e5f36e758449a862aa5272cb",
        "cbb5af185e942a2402f9eacbc0ed5bb876eea3c1223623d00447e4f3ba554b65",
        "cf56ff46a4a186109dd96584b5eeb58a510c4275b0e5f94f40bbae865e19f673",
        "d17cd8ecd586b712238a482ce46fa5293970742f276d8ab6a9e46ee0288f3355",
        "ddcef1660de3b06996507f56168865a20b43cda89cf7e8735a82b83bba00c498",
        "e389360d0fdbaeb3d250584b4730314e222f39c156a020144e8d960561791506",
        "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
        "e6b8f8766485f807ae7f8dac1670461f07c0a13eef3a1ff717538d7abad391b4",
        "eb04cf5eb1f39afa762f2bb120f296cba520c1b97db1589565b81cb9a17b7244",
        "ebf3c02a8789b1fb7d511995d663b72906d913ce0d5e10568a8a77e2586167e7",
        "f5074a8f5b9a5b8142f34abe152f60364d770eae75ee3eeceb45b6b996509788",
        "f59db3f45d57fcec94ccd516e6c8ccb20dd4363feb2c44d8656e95f50fdd8df8",
        "fe863d0822fe7a2353fa484d5924e875656d3dc9fb58771f6f616f9d571bc592",
        "ff856a2d251dcd88d36656f450126798cfabaade40799c722de4d2b5db36a73a",
    ]

    return root_hash in roots


def get_ct_log_name(log_id: str):
    global _ct_log_info

    # check to see if we have the CT log info cached
    if len(_ct_log_info) == 0:
        _ct_log_info = _get_ct_log_data()

    if _ct_log_info.get(log_id) is not None:
        return _ct_log_info[log_id]
    else:
        return "(Unknown: {hash})".format(hash=log_id)


def _get_ct_log_data():
    ct_log_info = {}

    data, _ = network.http_json(
        "https://www.gstatic.com/ct/log_list/all_logs_list.json"
    )

    for log in data["logs"]:
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(base64.b64decode(log["key"]))
        final = digest.finalize()
        value = bytes.hex(final)

        ct_log_info[value] = log["description"]

    return ct_log_info
